#include "magic_probability_toolkit.h"
#include "table_output.h"

using namespace mpt;
using namespace tbo;

void article_02()
{
	{
		CTable table( "Chance of drawing 3 or more spells in your first 4 draws for various starting land counts in a 7-card hand." );
		table.AddColumn( "land" );
		table.AddColumn( "probability" );

		for ( int lands_in_starting_hand = 2; lands_in_starting_hand < 8; ++lands_in_starting_hand )
		{
			CProbabilityDistribution state( 40, 17, 7, lands_in_starting_hand );
			state.DrawCards(4);

			const int spells_in_starting_hand = 7 - lands_in_starting_hand;
			const double probability = state.TakeMeasurement(
				[ spells_in_starting_hand ]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_spells >= spells_in_starting_hand + 3;
				}
			);
			table.AddInteger( lands_in_starting_hand );
			table.AddProbability( probability );
		}

		table.Print();
	}

	const int deck_sizes[] = { 40, 60, 100 };
	for ( int deck_size_idx = 0; deck_size_idx < std::extent< decltype( deck_sizes ) >::value; ++deck_size_idx )
	{
		const int deck_size = deck_sizes[ deck_size_idx ];

		const int significant_card_advantage = 2;
		const int turn_of_player_on_the_draw = 7;
		const int lands_in_deck = int( floor( 17.0 / 40.0 * deck_size ) );
		
		CTable average_table( "Average card advantage that my opponent has during the player on the draw's %dth turn when I start the game with a given number of lands and my opponent starts with 3 lands.  Our decks contain %d cards with %d lands.", turn_of_player_on_the_draw, deck_size, lands_in_deck );
		average_table.AddColumn( "land" );
		average_table.AddColumn( "card advantage" );

		CTable probability_table( "Probability of my opponent not having a card advantage during the player on the draw's %dth turn when I start the game with a given number of lands and my opponent starts with 3 lands.  Our decks contain %d cards with %d lands.", turn_of_player_on_the_draw, deck_size, lands_in_deck );
		probability_table.AddColumn( "land" );
		probability_table.AddColumn( "probability" );

		for ( int lands_in_starting_hand = 2; lands_in_starting_hand < 8; ++lands_in_starting_hand )
		{
			probability_table.AddInteger( lands_in_starting_hand );
			average_table.AddInteger( lands_in_starting_hand );

			double probability_of_no_card_disadvantage_on_the_draw;
			double average_card_disadvantage_on_the_draw;
			{
				CProbabilityDistribution me_state( deck_size, lands_in_deck, 7, lands_in_starting_hand );
				CProbabilityDistribution op_state( deck_size, lands_in_deck, 7, 3 );

				me_state.DrawCards( turn_of_player_on_the_draw );
				op_state.DrawCards( turn_of_player_on_the_draw - 1 );

				probability_of_no_card_disadvantage_on_the_draw = me_state.TakeMeasurement(
					[ op_state ]( const CProbabilityDistributionEntry& entry ) -> double
					{
						const int me_spell_count = entry.m_player.m_hand.m_spells;

						return op_state.TakeMeasurement(
							[ me_spell_count ]( const CProbabilityDistributionEntry& entry )
							{
								return entry.m_player.m_hand.m_spells - me_spell_count <= 0;
							}
						);
					}
				);

				average_card_disadvantage_on_the_draw = me_state.TakeMeasurement(
					[ op_state ]( const CProbabilityDistributionEntry& entry ) -> double
					{
						const int me_spell_count = entry.m_player.m_hand.m_spells;

						return op_state.TakeMeasurement(
							[ me_spell_count ]( const CProbabilityDistributionEntry& entry )
							{
								return entry.m_player.m_hand.m_spells - me_spell_count;
							}
						);
					}
				);
			}

			double probability_of_no_card_disadvantage_on_the_play;
			double average_card_disadvantage_on_the_play;
			{
				CProbabilityDistribution me_state( deck_size, lands_in_deck, 7, lands_in_starting_hand );
				CProbabilityDistribution op_state( deck_size, lands_in_deck, 7, 3 );

				me_state.DrawCards( turn_of_player_on_the_draw - 1 );
				op_state.DrawCards( turn_of_player_on_the_draw );

				probability_of_no_card_disadvantage_on_the_play = me_state.TakeMeasurement(
					[ op_state ]( const CProbabilityDistributionEntry& entry ) -> double
					{
						const int me_spell_count = entry.m_player.m_hand.m_spells;

						return op_state.TakeMeasurement(
							[ me_spell_count ]( const CProbabilityDistributionEntry& entry )
							{
								return entry.m_player.m_hand.m_spells - me_spell_count <= 0;
							}
						);
					}
				);

				average_card_disadvantage_on_the_play = me_state.TakeMeasurement(
					[ op_state ]( const CProbabilityDistributionEntry& entry ) -> double
					{
						const int me_spell_count = entry.m_player.m_hand.m_spells;

						return op_state.TakeMeasurement(
							[ me_spell_count ]( const CProbabilityDistributionEntry& entry )
							{
								return entry.m_player.m_hand.m_spells - me_spell_count;
							}
						);
					}
				);
			}

			const double probability_of_no_card_disadvantage_average = ( probability_of_no_card_disadvantage_on_the_play + probability_of_no_card_disadvantage_on_the_draw ) / 2.0;
			const double average_card_disadvantage_average = ( average_card_disadvantage_on_the_play + average_card_disadvantage_on_the_draw ) / 2.0;

			average_table.AddNumber( average_card_disadvantage_average );
			probability_table.AddProbability( probability_of_no_card_disadvantage_average );
		}

		average_table.Print();
		probability_table.Print();
	}
}