#include "magic_probability_toolkit.h"
#include "table_output.h"
#include "mulligan_rules.h"
#include "turn_cost.h"

using namespace mpt;
using namespace tbo;
using namespace mlr;
using namespace ctc;

struct CTurnCostListEntry
{
	int m_casting_cost;
	int m_colored_mana_C;
	int m_colored_mana_D;
	double m_turn_cost;
};

enum
{
	CastingCostToStringBufferSize = 16
};

void casting_cost_to_string( char buffer[ CastingCostToStringBufferSize ], int converted_mana_cost, int C_cost, int D_cost )
{
	const int colored_mana_count = C_cost + D_cost;

	assert( converted_mana_cost >= colored_mana_count );

	int next_character_index = 0;
	if ( converted_mana_cost > colored_mana_count )
	{
		next_character_index += sprintf_s( &buffer[ next_character_index ], CastingCostToStringBufferSize - next_character_index, "%d", converted_mana_cost - colored_mana_count );
	}
	while ( C_cost > 0 )
	{
		buffer[ next_character_index ] = 'C';
		++next_character_index;
		--C_cost;
	}
	while ( D_cost > 0 )
	{
		buffer[ next_character_index ] = 'D';
		++next_character_index;
		--D_cost;
	}

	buffer[ next_character_index ] = '\0';
	++next_character_index;

	assert( next_character_index <= CastingCostToStringBufferSize );
}

void article_14()
{
	{
		typedef CardSet::TwoLandColorsWithFixing Card;
		typedef CardSetTypes< Card >::CPile CPile;
		typedef CardSetTypes< Card >::CProbabilityDistributionEntry CProbabilityDistributionEntry;
		typedef CardSetTypes< Card >::CProbabilityDistribution CProbabilityDistribution;

		auto initial_hand_evaluation = []( CProbabilityDistributionEntry& entry )
		{
			typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType TCardSetType;
			if ( entry.m_player.m_hand.m_cards[ TCardSetType::LandC ] + entry.m_player.m_hand.m_cards[ TCardSetType::Fixing ] < 1 || entry.m_player.m_hand.m_cards[ TCardSetType::LandD ] + entry.m_player.m_hand.m_cards[ TCardSetType::Fixing ] < 1  )
			{
				entry.m_user_data = 1.0;
			}
		};

		auto competitive_start_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
		{
			typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType TCardSetType;

			if ( entry.m_player.m_hand.GetSize() == 7 && entry.m_user_data != 0.0 ) return false;

			const int mana_sources = entry.m_player.m_hand.m_cards[ TCardSetType::LandC ] + entry.m_player.m_hand.m_cards[ TCardSetType::LandD ] + entry.m_player.m_hand.m_cards[ TCardSetType::Fixing ];
			return mana_sources >= 4 && entry.m_player.m_hand.m_cards[ TCardSetType::Spell ] >= 4;
		};

		CPile deck;
		deck.m_cards[ Card::LandC ] = 8;
		deck.m_cards[ Card::LandD ] = 8;
		deck.m_cards[ Card::Fixing ] = 1;
		deck.m_cards[ Card::Spell ] = 23;
		deck.m_cards[ Card::Turn ] = 0;

		CMulliganRulesSet< Card > rules_set = CalculateMulliganRulesSet<
			CProbabilityDistribution,
			decltype( initial_hand_evaluation ),
			decltype( competitive_start_measurement )
		>(
			deck,
			initial_hand_evaluation,
			competitive_start_measurement
		);

		for ( int casting_cost = 1; casting_cost <= 6; ++casting_cost )
		{
			CTable table( "Turn cost for various casting costs of converted mana cost %d.", casting_cost );
			table.AddColumn( "casting cost" );
			table.AddColumn( "turn cost" );

			for ( int colored_mana = 0; colored_mana <= casting_cost; ++colored_mana )
			{
				{
					char buffer[ CastingCostToStringBufferSize ];
					casting_cost_to_string( buffer, casting_cost, colored_mana, 0 );
					table.AddString( buffer );
				}

				auto execute_turn = [ casting_cost, colored_mana ]( CProbabilityDistributionEntry& entry ) -> bool
				{
					typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType Card;

					const int turn = entry.m_player.m_hand.m_cards[ Card::Turn ];

					const int total_land_in_play = std::min( turn, entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::LandD ] + entry.m_player.m_hand.m_cards[ Card::Fixing ] );
					const int total_C_in_play = std::min( turn, entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::Fixing ] );

					if ( total_land_in_play >= casting_cost && total_C_in_play >= colored_mana )
					{
						return true;
					}

					++entry.m_player.m_hand.m_cards[ Card::Turn ];
					return false;
				};

				{
					const double average_turn_cost = CalculateTurnCost<
						CProbabilityDistribution,
						decltype( execute_turn )
					>( deck, execute_turn, rules_set );

					table.AddNumber( average_turn_cost, 2 );
				}
			}

			table.Print();
		}

		for ( int casting_cost = 1; casting_cost <= 2; ++casting_cost )
		{
			const int colored_mana = casting_cost;

			char buffer[ CastingCostToStringBufferSize ];
			casting_cost_to_string( buffer, casting_cost, casting_cost, 0 );

			CTable table( "Casting turn distribution of a spell costing %s.", buffer );
			table.AddColumn( "turn" );
			table.AddColumn( "probability" );

			auto execute_turn = [ casting_cost, colored_mana ]( CProbabilityDistributionEntry& entry ) -> bool
			{
				typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType Card;

				const int turn = entry.m_player.m_hand.m_cards[ Card::Turn ];

				const int total_land_in_play = std::min( turn, entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::LandD ] + entry.m_player.m_hand.m_cards[ Card::Fixing ] );
				const int total_C_in_play = std::min( turn, entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::Fixing ] );

				if ( total_land_in_play >= casting_cost && total_C_in_play >= colored_mana )
				{
					return true;
				}

				++entry.m_player.m_hand.m_cards[ Card::Turn ];
				return false;
			};

			auto get_turn = []( const CProbabilityDistributionEntry& entry ) -> int
			{
				typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType Card;
				return entry.m_player.m_hand.m_cards[ Card::Turn ];
			};

			auto set_first_turn = []( CProbabilityDistributionEntry& entry )
			{
				typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType Card;
				entry.m_player.m_hand.m_cards[ Card::Turn ] = 1;
			};

			{
				double casting_turn_on_the_play_table[ 100 ];
				{
					CProbabilityDistribution state( deck );
					state.DrawStartingHandFromRules( rules_set.m_on_the_play );

					state.ForEach( set_first_turn );

					state.TakeTurnsUntil( execute_turn );
					const double turn_cost = state.TakeMeasurement( get_turn );

					state.BinDistribution(
						casting_turn_on_the_play_table,
						std::extent< decltype( casting_turn_on_the_play_table ) >::value,
						get_turn
					);
				}

				double casting_turn_on_the_draw_table[ 100 ];
				{
					CProbabilityDistribution state( deck );
					state.DrawStartingHandFromRules( rules_set.m_on_the_draw );

					state.ForEach( set_first_turn );
					state.DrawCard();

					state.TakeTurnsUntil( execute_turn );
					const double turn_cost = state.TakeMeasurement( get_turn );

					state.BinDistribution(
						casting_turn_on_the_draw_table,
						std::extent< decltype( casting_turn_on_the_draw_table ) >::value,
						get_turn
					);
				}

				double average_casting_turn_if_not_first_castable_turn = 0.0;
				for ( int turn = 0; turn < std::extent< decltype( casting_turn_on_the_draw_table ) >::value; ++turn )
				{
					const double probability = ( casting_turn_on_the_play_table[ turn ] + casting_turn_on_the_draw_table[ turn ] ) / 2.0;
					if ( probability > 0.0 )
					{
						table.AddInteger( turn );
						table.AddProbability( probability, 2 );

						if ( turn > casting_cost )
						{
							average_casting_turn_if_not_first_castable_turn += probability * turn;
						}
					}
				}

				const double probability_of_casting_after_first_castable_turn = 1.0 - ( casting_turn_on_the_play_table[ casting_cost ] + casting_turn_on_the_draw_table[ casting_cost ] ) / 2.0;

				average_casting_turn_if_not_first_castable_turn /= probability_of_casting_after_first_castable_turn;

				table.AddAddendum( "probability of not casting on first castable turn" );
				table.AddProbability( probability_of_casting_after_first_castable_turn );

				table.AddAddendum( "average casting turn when not cast on first castable turn" );
				table.AddNumber( average_casting_turn_if_not_first_castable_turn );

				table.AddAddendum( "average casting turn minus first castable turn" );
				table.AddNumber( ( average_casting_turn_if_not_first_castable_turn - casting_cost ) * probability_of_casting_after_first_castable_turn, 2 );

				table.Print();
			}
		}
	}

	double turn_cost_of_single_colored_mana_spells_table[ 11 ];

	{
		typedef CardSet::TwoLandColorsWithFixing Card;
		typedef CardSetTypes< Card >::CPile CPile;
		typedef CardSetTypes< Card >::CProbabilityDistributionEntry CProbabilityDistributionEntry;
		typedef CardSetTypes< Card >::CProbabilityDistribution CProbabilityDistribution;

		CTable table( "Turn cost for various casting costs." );
		table.AddColumn( "casting cost" );
		table.AddColumn( "turn cost" );

		auto initial_hand_evaluation = []( CProbabilityDistributionEntry& entry )
		{
			typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType TCardSetType;
			if ( entry.m_player.m_hand.m_cards[ TCardSetType::LandC ] + entry.m_player.m_hand.m_cards[ TCardSetType::Fixing ] < 1 || entry.m_player.m_hand.m_cards[ TCardSetType::LandD ] + entry.m_player.m_hand.m_cards[ TCardSetType::Fixing ] < 1  )
			{
				entry.m_user_data = 1.0;
			}
		};

		auto competitive_start_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
		{
			typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType TCardSetType;

			if ( entry.m_player.m_hand.GetSize() == 7 && entry.m_user_data != 0.0 ) { return false; }

			const int mana_sources = entry.m_player.m_hand.m_cards[ TCardSetType::LandC ] + entry.m_player.m_hand.m_cards[ TCardSetType::LandD ] + entry.m_player.m_hand.m_cards[ TCardSetType::Fixing ];
			return mana_sources >= 4 && entry.m_player.m_hand.m_cards[ TCardSetType::Spell ] >= 4;
		};

		std::list< CTurnCostListEntry > turn_costs;

		CPile deck;
		deck.m_cards[ Card::LandC ] = 8;
		deck.m_cards[ Card::LandD ] = 8;
		deck.m_cards[ Card::Fixing ] = 1;
		deck.m_cards[ Card::Spell ] = 23;
		deck.m_cards[ Card::Turn ] = 0;

		CMulliganRulesSet< Card > rules_set = CalculateMulliganRulesSet<
			CProbabilityDistribution,
			decltype( initial_hand_evaluation ),
			decltype( competitive_start_measurement )
		>(
			deck,
			initial_hand_evaluation,
			competitive_start_measurement
		);

		for ( int casting_cost = 1; casting_cost <= 11; ++casting_cost )
		{
			for ( int colored_mana_C = 0; colored_mana_C <= std::min( casting_cost, 4 ); ++colored_mana_C )
			{
				for ( int colored_mana_D = 0; colored_mana_D <= colored_mana_C && colored_mana_D + colored_mana_C <= std::min( casting_cost, 4 ); ++colored_mana_D )
				{
					auto execute_turn = [ casting_cost, colored_mana_C, colored_mana_D ]( CProbabilityDistributionEntry& entry ) -> bool
					{
						typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType Card;

						const int turn = entry.m_player.m_hand.m_cards[ Card::Turn ];
						const int total_land_in_play = std::min( turn, entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::LandD ] + entry.m_player.m_hand.m_cards[ Card::Fixing ] );
						const int C_land_in_play = std::min( turn, entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::Fixing ] );
						const int D_land_in_play = std::min( turn, entry.m_player.m_hand.m_cards[ Card::LandD ] + entry.m_player.m_hand.m_cards[ Card::Fixing ] );

						if ( total_land_in_play >= casting_cost && C_land_in_play >= colored_mana_C && D_land_in_play >= colored_mana_D )
						{
							return true;
						}

						++entry.m_player.m_hand.m_cards[ Card::Turn ];
						return false;
					};

					const double average_turn_cost = CalculateTurnCost<
						CProbabilityDistribution,
						decltype( execute_turn )
					>( deck, execute_turn, rules_set );

					CTurnCostListEntry new_entry = {
						casting_cost,
						colored_mana_C,
						colored_mana_D,
						average_turn_cost
					};

					turn_costs.push_back( new_entry );

					if ( colored_mana_C == 1 && colored_mana_D == 0 && casting_cost < std::extent< decltype( turn_cost_of_single_colored_mana_spells_table ) >::value )
					{
						turn_cost_of_single_colored_mana_spells_table[ casting_cost ] = average_turn_cost;
					}
				}
			}
		}

		turn_costs.sort( 
			[]( const CTurnCostListEntry& entry_a, const CTurnCostListEntry& entry_b )
			{
				return entry_a.m_turn_cost < entry_b.m_turn_cost;
			}
		);

		for ( auto i = turn_costs.begin(); i != turn_costs.end(); ++i )
		{
			char buffer[ CastingCostToStringBufferSize ];
			casting_cost_to_string( buffer, i->m_casting_cost, i->m_colored_mana_C, i->m_colored_mana_D );

			table.AddString( buffer );
			table.AddNumber( i->m_turn_cost, 2 );
		}

		table.Print();
	}

	{
		typedef CardSet::TwoLandColorsWithFixing Card;
		typedef CardSetTypes< Card >::CPile CPile;
		typedef CardSetTypes< Card >::CProbabilityDistributionEntry CProbabilityDistributionEntry;
		typedef CardSetTypes< Card >::CProbabilityDistribution CProbabilityDistribution;

		auto initial_hand_evaluation = []( CProbabilityDistributionEntry& entry ) {};

		auto competitive_start_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
		{
			typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType TCardSetType;

			const int mana_sources = entry.m_player.m_hand.m_cards[ TCardSetType::LandC ] + entry.m_player.m_hand.m_cards[ TCardSetType::LandD ] + entry.m_player.m_hand.m_cards[ TCardSetType::Fixing ];
			return mana_sources >= 4 && entry.m_player.m_hand.m_cards[ TCardSetType::Spell ] >= 4;
		};

		const int low_casting_cost = 2;
		const int high_casting_cost = 6;

		CTable table_crossover( "Number of C-lands in deck when two casting costs have equal turn cost." );
		table_crossover.AddColumn( "converted mana cost" );
		table_crossover.AddColumn( "casting costs" );
		table_crossover.AddColumn( "crossover point" );

		for ( int casting_cost = low_casting_cost; casting_cost <= high_casting_cost; ++casting_cost )
		{
			CTable table_turn_cost( "Turn costs when deck has an asymmetric mana base." );
			table_turn_cost.AddColumn( "C-land count" );

			table_crossover.AddInteger( casting_cost );

			{
				char buffer_CC[ CastingCostToStringBufferSize ];
				casting_cost_to_string( buffer_CC, casting_cost, 2, 0 );

				char buffer_D[ CastingCostToStringBufferSize ];
				casting_cost_to_string( buffer_D, casting_cost, 0, 1 );

				table_turn_cost.AddColumn( buffer_CC );
				table_turn_cost.AddColumn( buffer_D );

				table_crossover.AddString( "%s and %s", buffer_CC, buffer_D );
			}

			const int low_land_count = 8;
			const int high_land_count = 12;

			double results_table[ 2 ][ 1 + high_land_count - low_land_count ];

			for ( int C_land_count = low_land_count; C_land_count <= high_land_count; ++C_land_count )
			{
				table_turn_cost.AddInteger( C_land_count );

				for ( int colored_mana = 2; colored_mana >= 1; --colored_mana )
				{
					CPile deck;
					if ( colored_mana == 2 )
					{
						deck.m_cards[ Card::LandC ] = C_land_count;
						deck.m_cards[ Card::LandD ] = 16 - C_land_count;
					}
					else
					{
						deck.m_cards[ Card::LandC ] = 16 - C_land_count;
						deck.m_cards[ Card::LandD ] = C_land_count;
					}
					deck.m_cards[ Card::Fixing ] = 1;
					deck.m_cards[ Card::Spell ] = 23;
					deck.m_cards[ Card::Turn ] = 0;

					CMulliganRulesSet< Card > rules_set = CalculateMulliganRulesSet<
						CProbabilityDistribution,
						decltype( initial_hand_evaluation ),
						decltype( competitive_start_measurement )
					>(
						deck,
						initial_hand_evaluation,
						competitive_start_measurement
					);

					auto execute_turn = [ casting_cost, colored_mana ]( CProbabilityDistributionEntry& entry ) -> bool
					{
						typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType Card;

						const int turn = entry.m_player.m_hand.m_cards[ Card::Turn ];

						const int total_land_in_play = std::min( turn, entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::LandD ] + entry.m_player.m_hand.m_cards[ Card::Fixing ] );
						const int total_C_in_play = std::min( turn, entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::Fixing ] );

						if ( total_land_in_play >= casting_cost && total_C_in_play >= colored_mana )
						{
							return true;
						}

						++entry.m_player.m_hand.m_cards[ Card::Turn ];
						return false;
					};

					{
						const double average_turn_cost = CalculateTurnCost<
							CProbabilityDistribution,
							decltype( execute_turn )
						>( deck, execute_turn, rules_set );

						results_table[ colored_mana - 1 ][ C_land_count - low_land_count ] = average_turn_cost;

						table_turn_cost.AddNumber( average_turn_cost, 2 );
					}
				}
			}

			table_turn_cost.AddAddendum( "crossover point" );

			for ( int C_land_count = low_land_count + 1; C_land_count <= high_land_count; ++C_land_count )
			{
				const int index = C_land_count - low_land_count;

				if ( results_table[ 0 ][ index ] > results_table[ 1 ][ index ] )
				{
					const double a = results_table[ 0 ][ index - 1 ] - results_table[ 1 ][ index - 1 ];
					const double b = results_table[ 1 ][ index ] - results_table[ 0 ][ index ];
					const double zero = a / ( a + b );

					const double crossover_point = C_land_count - 1 + zero;

					table_turn_cost.AddNumber( crossover_point );
					table_crossover.AddNumber( crossover_point );

					break;
				}
			}

			table_turn_cost.Print();
		}

		table_crossover.Print();
	}

	{
		typedef CardSet::TwoLandColors Card;
		typedef CardSetTypes< Card >::CPile CPile;
		typedef CardSetTypes< Card >::CProbabilityDistributionEntry CProbabilityDistributionEntry;
		typedef CardSetTypes< Card >::CProbabilityDistribution CProbabilityDistribution;

		CTable table( "Turn cost for a splashed 4C spell with various numbers of splash sources in deck." );
		table.AddColumn( "splash sources" );
		table.AddColumn( "turn cost" );

		auto initial_hand_evaluation = []( CProbabilityDistributionEntry& entry ) {};

		auto competitive_start_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
		{
			typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType TCardSetType;

			const int mana_sources = entry.m_player.m_hand.m_cards[ TCardSetType::LandC ] + entry.m_player.m_hand.m_cards[ TCardSetType::LandD ];
			return mana_sources >= 4 && entry.m_player.m_hand.m_cards[ TCardSetType::Spell ] >= 4;
		};

		const int casting_cost = 5;
		const int colored_mana = 1;

		for ( int num_splash_sources = 1; num_splash_sources <= 6; ++num_splash_sources )
		{
			table.AddInteger( num_splash_sources );

			auto execute_turn = [ casting_cost, colored_mana ]( CProbabilityDistributionEntry& entry ) -> bool
			{
				typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType Card;

				const int turn = entry.m_player.m_hand.m_cards[ Card::Turn ];

				const int total_land_in_play = std::min( turn, entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::LandD ] );
				const int total_C_in_play = std::min( turn, entry.m_player.m_hand.m_cards[ Card::LandC ] );

				if ( total_land_in_play >= casting_cost && total_C_in_play >= colored_mana )
				{
					return true;
				}

				++entry.m_player.m_hand.m_cards[ Card::Turn ];
				return false;
			};

			CPile deck;
			deck.m_cards[ Card::LandC ] = num_splash_sources;
			deck.m_cards[ Card::LandD ] = 17 - num_splash_sources;
			deck.m_cards[ Card::Spell ] = 23;
			deck.m_cards[ Card::Turn ] = 0;

			const double average_turn_cost = CalculateTurnCost<
				CProbabilityDistribution,
				decltype( initial_hand_evaluation ),
				decltype( competitive_start_measurement ),
				decltype( execute_turn )
			>( deck, initial_hand_evaluation, competitive_start_measurement, execute_turn );

			table.AddNumber( average_turn_cost );
		}

		table.Print();
	}

	{
		typedef CardSet::TwoLandColors Card;
		typedef CardSetTypes< Card >::CPile CPile;
		typedef CardSetTypes< Card >::CProbabilityDistributionEntry CProbabilityDistributionEntry;
		typedef CardSetTypes< Card >::CProbabilityDistribution CProbabilityDistribution;

		CTable table( "Turn cost for splashed spell of various casting costs with 3 splash sources in deck." );
		table.AddColumn( "casting cost" );
		table.AddColumn( "turn cost" );

		auto initial_hand_evaluation = []( CProbabilityDistributionEntry& entry )
		{
			return;
		};

		auto competitive_start_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
		{
			typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType TCardSetType;

			const int mana_sources = entry.m_player.m_hand.m_cards[ TCardSetType::LandC ] + entry.m_player.m_hand.m_cards[ TCardSetType::LandD ];
			return mana_sources >= 4 && entry.m_player.m_hand.m_cards[ TCardSetType::Spell ] >= 4;
		};

		std::list< CTurnCostListEntry > turn_costs;

		const int num_splash_sources = 3;

		for ( int casting_cost = 1; casting_cost <= 7; ++casting_cost )
		{
			for ( int colored_mana = 1; colored_mana <= std::min( casting_cost, num_splash_sources ); ++colored_mana )
			{
				auto execute_turn = [ casting_cost, colored_mana ]( CProbabilityDistributionEntry& entry ) -> bool
				{
					typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType Card;

					const int turn = entry.m_player.m_hand.m_cards[ Card::Turn ];

					const int total_land_in_play = std::min( turn, entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::LandD ] );
					const int total_C_in_play = std::min( turn, entry.m_player.m_hand.m_cards[ Card::LandC ] );

					if ( total_land_in_play >= casting_cost && total_C_in_play >= colored_mana )
					{
						return true;
					}

					++entry.m_player.m_hand.m_cards[ Card::Turn ];
					return false;
				};

				CPile deck;
				deck.m_cards[ Card::LandC ] = num_splash_sources;
				deck.m_cards[ Card::LandD ] = 17 - num_splash_sources;
				deck.m_cards[ Card::Spell ] = 23;
				deck.m_cards[ Card::Turn ] = 0;

				const double average_turn_cost = CalculateTurnCost<
					CProbabilityDistribution,
					decltype( initial_hand_evaluation ),
					decltype( competitive_start_measurement ),
					decltype( execute_turn )
				>( deck, initial_hand_evaluation, competitive_start_measurement, execute_turn );

				CTurnCostListEntry new_entry = {
					casting_cost,
					colored_mana,
					0,
					average_turn_cost
				};

				turn_costs.push_back( new_entry );
			}
		}

		turn_costs.sort( 
			[]( const CTurnCostListEntry& entry_a, const CTurnCostListEntry& entry_b )
			{
				return entry_a.m_turn_cost < entry_b.m_turn_cost;
			}
		);

		for ( auto i = turn_costs.begin(); i != turn_costs.end(); ++i )
		{
			char buffer[ CastingCostToStringBufferSize ];
			casting_cost_to_string( buffer, i->m_casting_cost, i->m_colored_mana_C, i->m_colored_mana_D );

			table.AddString( buffer );
			table.AddNumber( i->m_turn_cost );
		}

		table.Print();
	}

	{
		typedef CardSet::TwoLandColors Card;
		typedef CardSetTypes< Card >::CPile CPile;
		typedef CardSetTypes< Card >::CProbabilityDistributionEntry CProbabilityDistributionEntry;
		typedef CardSetTypes< Card >::CProbabilityDistribution CProbabilityDistribution;

		CTable table( "Turn cost and effective converted mana cost for splashed spell of various casting costs with 3 splash sources in deck." );
		table.AddColumn( "casting cost" );
		table.AddColumn( "splashed" );
		table.AddColumn( "normal" );
		table.AddColumn( "turn cost premium" );
		table.AddColumn( "effective CMC" );

		auto initial_hand_evaluation = []( CProbabilityDistributionEntry& entry )
		{
			return;
		};

		auto competitive_start_measurement = []( const CProbabilityDistributionEntry& entry ) -> bool
		{
			typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType TCardSetType;

			const int mana_sources = entry.m_player.m_hand.m_cards[ TCardSetType::LandC ] + entry.m_player.m_hand.m_cards[ TCardSetType::LandD ];
			return mana_sources >= 4 && entry.m_player.m_hand.m_cards[ TCardSetType::Spell ] >= 4;
		};

		std::list< CTurnCostListEntry > turn_costs;

		const int num_splash_sources = 3;
		const int colored_mana = 1;

		for ( int casting_cost = 1; casting_cost <= 7; ++casting_cost )
		{
			{
				char buffer[ CastingCostToStringBufferSize ];
				casting_cost_to_string( buffer, casting_cost, colored_mana, 0 );
				table.AddString( buffer );
			}

			auto execute_turn = [ casting_cost, colored_mana ]( CProbabilityDistributionEntry& entry ) -> bool
			{
				typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType Card;

				const int turn = entry.m_player.m_hand.m_cards[ Card::Turn ];

				const int total_land_in_play = std::min( turn, entry.m_player.m_hand.m_cards[ Card::LandC ] + entry.m_player.m_hand.m_cards[ Card::LandD ] );
				const int total_C_in_play = std::min( turn, entry.m_player.m_hand.m_cards[ Card::LandC ] );

				if ( total_land_in_play >= casting_cost && total_C_in_play >= colored_mana )
				{
					return true;
				}

				++entry.m_player.m_hand.m_cards[ Card::Turn ];
				return false;
			};

			double average_turn_cost_splashed;
			{
				CPile deck;
				deck.m_cards[ Card::LandC ] = num_splash_sources;
				deck.m_cards[ Card::LandD ] = 17 - num_splash_sources;
				deck.m_cards[ Card::Spell ] = 23;
				deck.m_cards[ Card::Turn ] = 0;

				average_turn_cost_splashed = CalculateTurnCost<
					CProbabilityDistribution,
					decltype( initial_hand_evaluation ),
					decltype( competitive_start_measurement ),
					decltype( execute_turn )
				>( deck, initial_hand_evaluation, competitive_start_measurement, execute_turn );

				table.AddNumber( average_turn_cost_splashed );
			}

			assert( casting_cost < std::extent< decltype( turn_cost_of_single_colored_mana_spells_table ) >::value );
			const double average_turn_cost_normal = turn_cost_of_single_colored_mana_spells_table[ casting_cost ];
			table.AddNumber( average_turn_cost_normal );

			table.AddNumber( average_turn_cost_splashed - average_turn_cost_normal );

			for ( int cost = 1; cost < std::extent< decltype( turn_cost_of_single_colored_mana_spells_table ) >::value; ++cost )
			{
				if ( turn_cost_of_single_colored_mana_spells_table[ cost ] > average_turn_cost_splashed )
				{
					const double a = turn_cost_of_single_colored_mana_spells_table[ cost ] - turn_cost_of_single_colored_mana_spells_table[ cost - 1 ];
					const double b = average_turn_cost_splashed - turn_cost_of_single_colored_mana_spells_table[ cost - 1 ];
					const double result = cost - 1 + b / a;

					table.AddNumber( result );

					break;
				}

				if ( cost == std::extent< decltype( turn_cost_of_single_colored_mana_spells_table ) >::value - 1 )
				{
					table.AddString( "X" );
				}
			}
		}

		table.Print();
	}
}