#include <random>

#include "magic_probability_toolkit.h"
#include "mulligan_rules.h"
#include "table_output.h"

using namespace mpt;
using namespace tbo;
using namespace mlr;

namespace A15_P01
{

static const bool VERBOSE = true;

static const int DECK_SIZE = 40;
static const int LAND_COUNT = 17;

std::default_random_engine random( 1 );

typedef CardSet::ManaCost Card;
typedef CardSetTypes< Card >::CPile CPile;
typedef CardSetTypes< Card >::CProbabilityDistributionEntry CProbabilityDistributionEntry;
typedef CardSetTypes< Card >::CProbabilityDistribution CProbabilityDistribution;

struct CDeck
{
	CPile deck;
	double value;
};

typedef std::vector< CDeck > CDeckCache;

enum TPlayOrDrawScheme
{
	OnThePlay,
	OnTheDraw,
	Averaged,
	LastCalculablePlayOrDrawValue = Averaged,
	Averaged_UsePlayOrDrawShortcut,
};

struct CSituation
{
	int number_of_turns;
	TPlayOrDrawScheme play_or_draw_scheme;
	int locked_land_count;
	bool use_mulligan_shortcut;
	int number_of_dead_cards;
};

void print_deck( FILE* p_file, const CPile& deck )
{
	fprintf( p_file, "[L:%.2d ", deck.m_cards[ Card::Land ] );
	for ( int casting_cost = Card::ManaCost_1; casting_cost < Card::DeadCard; ++casting_cost )
	{
		fprintf(
			p_file,
			"%d:%.2d ",
			casting_cost,
			deck.m_cards[ casting_cost ]
		);
	}
	fprintf( p_file, "D:%.2d]", deck.m_cards[ Card::DeadCard ] );
}

void print_iteration( FILE* p_file, int iteration, bool cache_hit, const CDeck& location, bool result )
{
	fprintf( stderr, "         %.3d[%c]: ", iteration, cache_hit ? 'c' : '+' );
	print_deck( stderr, location.deck );
	fprintf( stderr, ": m = %.4f%s\n", location.value, result ? " -> keep" : "" );
	fflush( stderr );
}

void print_deck_table( const char* header, int header_data, const CDeck& deck, const char* addendum = NULL, const char* extra_addendum = NULL, double extra_addendum_value = 0.0 )
{
	CTable table( header, header_data );
	table.AddColumn( "" );
	table.AddColumn( "land" );
	table.AddColumn( "CC1" );
	table.AddColumn( "CC2" );
	table.AddColumn( "CC3" );
	table.AddColumn( "CC4" );
	table.AddColumn( "CC5" );

	table.AddString( "maximizing deck" );
	table.AddInteger( deck.deck.m_cards[ Card::Land ] );
	table.AddInteger( deck.deck.m_cards[ Card::ManaCost_1 ] );
	table.AddInteger( deck.deck.m_cards[ Card::ManaCost_2 ] );
	table.AddInteger( deck.deck.m_cards[ Card::ManaCost_3 ] );
	table.AddInteger( deck.deck.m_cards[ Card::ManaCost_4 ] );
	table.AddInteger( deck.deck.m_cards[ Card::ManaCost_5 ] );

	if ( addendum )
	{
		table.AddAddendum( addendum );
		if ( extra_addendum )
		{
			table.AddProbability( 1.0 + deck.value, 1 );
		}
		else
		{
			table.AddNumber( deck.value, 2 );
		}
	}

	if ( extra_addendum )
	{
		table.AddAddendum( extra_addendum );
		table.AddNumber( extra_addendum_value, 2 );
	}

	table.Print();
}

template< typename TMulliganCondition, typename TMetric >
double calculate_metric_for_deck( const CPile& deck, const CSituation& situation, TMulliganCondition mulligan_condition, TMetric metric )
{
	assert( situation.play_or_draw_scheme <= LastCalculablePlayOrDrawValue );
	if ( situation.play_or_draw_scheme > LastCalculablePlayOrDrawValue )
	{
		printf( " Assert \"situation.play_or_draw_scheme <= LastCalculablePlayOrDrawValue\" failed.\n" );
		printf( " Exiting.\n" );
		exit( -1 );
	}

	double result_otp;
	if ( situation.play_or_draw_scheme == OnThePlay || situation.play_or_draw_scheme == Averaged )
	{
		CProbabilityDistribution state( deck );
		state.DrawStartingHand(
			[ mulligan_condition ]( const CProbabilityDistributionEntry& entry )
			{
				return mulligan_condition( Play, entry );
			}
		);
		for ( int turn = 1; turn <= situation.number_of_turns; ++turn )
		{
			if ( turn > 1 )
			{
				state.DrawCard();
			}
			state.ForEach( [ metric, turn ]( CProbabilityDistributionEntry& entry ) { return metric( turn, entry ); } );
		}
		result_otp = state.TakeMeasurement( []( const CProbabilityDistributionEntry& entry ) { return entry.m_user_data; } );
	}

	double result_otd;
	if ( situation.play_or_draw_scheme == OnTheDraw || situation.play_or_draw_scheme == Averaged )
	{
		CProbabilityDistribution state( deck );
		state.DrawStartingHand(
			[ mulligan_condition ]( const CProbabilityDistributionEntry& entry )
			{
				return mulligan_condition( Draw, entry );
			}
		);
		for ( int turn = 1; turn <= situation.number_of_turns; ++turn )
		{
			state.DrawCard();
			state.ForEach( [ metric, turn ]( CProbabilityDistributionEntry& entry ) { return metric( turn, entry ); } );
		}
		result_otd = state.TakeMeasurement( []( const CProbabilityDistributionEntry& entry ) { return entry.m_user_data; } );
	}

	switch ( situation.play_or_draw_scheme )
	{
	case OnThePlay:
		return result_otp;
	case OnTheDraw:
		return result_otd;
	case Averaged:
	default:
		return ( result_otp + result_otd ) / 2.0;
	}
}

template< typename TMulliganCondition, typename TMetric >
double determine_metric_for_deck( const CPile& deck, const CSituation& situation, TMulliganCondition mulligan_condition, TMetric metric, CDeckCache& cache )
{
	for ( auto i = cache.begin(); i != cache.end(); ++i )
	{
		if ( i->deck == deck )
		{
			return i->value;
		}
	}

	double value = calculate_metric_for_deck( deck, situation, mulligan_condition, metric );

	CDeck location = { deck, value };
	cache.push_back( location );

	return value;
}

template< typename TMulliganCondition, typename TMetric >
CDeck hill_climb( CPile initial_deck, const CSituation& situation, TMulliganCondition mulligan_condition, bool mulligan_condition_is_not_null, TMetric metric, CDeckCache& cache )
{
	if ( VERBOSE ) fprintf( stderr, "       Setting up hill climb algorithm.\n" );

	const int deck_size = DECK_SIZE;
	const int num_iterations = 999;

	bool land_count_is_locked = situation.locked_land_count != -1;
	assert( !land_count_is_locked || initial_deck.m_cards[ Card::Land ] == situation.locked_land_count );

	if ( VERBOSE )
	{
		if ( land_count_is_locked )
		{
			fprintf( stderr, "       Deck land count is locked at %d.\n", situation.locked_land_count );
		}
		else
		{
			fprintf( stderr, "       Deck land count is not locked.\n", situation.locked_land_count );
		}
		fprintf( stderr, "       Number of dead cards is set to %d.\n", situation.number_of_dead_cards );
		fprintf( stderr, "       Number of game turns is set to %d.\n", situation.number_of_turns );
		switch ( situation.play_or_draw_scheme )
		{
		case OnThePlay:
			fprintf( stderr, "       Simulating on the play.\n" );
			break;
		case OnTheDraw:
			fprintf( stderr, "       Simulating on the draw.\n" );
			break;
		case Averaged:
			fprintf( stderr, "       Simulating both on the play and on the draw and averaging.\n" );
			break;
		default:
			assert( false );
		}
		if ( mulligan_condition_is_not_null )
		{
			fprintf( stderr, "       Simulating with mulligans.\n" );
		}
		else
		{
			fprintf( stderr, "       Simulating without mulligans.\n" );
		}
		fprintf( stderr, "       Initial deck: " );
		print_deck( stderr, initial_deck );
		fprintf( stderr, ".\n" );
		fprintf( stderr, "       Starting hill climb.\n" );
	}
	
	int cache_size = cache.size();

	assert( initial_deck.GetSize() == deck_size );
	CDeck current = { initial_deck, determine_metric_for_deck( initial_deck, situation, mulligan_condition, metric, cache ) };

	if ( VERBOSE )
	{
		print_iteration( stderr, 0, cache_size == cache.size(), current, false );
	}

	bool tested_directions[ Card::NumberOfCurveRelevantCardTypes ][ Card::NumberOfCurveRelevantCardTypes ];
	memset( tested_directions, 0, sizeof( tested_directions ) );
	int try_next_increment = -1;
	int try_next_decrement = -1;
	int iteration = 0;

	if ( !land_count_is_locked || current.deck.m_cards[ Card::Land ] != DECK_SIZE )
	{
		while ( iteration < num_iterations )
		{
			int increment_random_axis = random() % Card::NumberOfCurveRelevantCardTypes;
			int decrement_random_axis = random() % Card::NumberOfCurveRelevantCardTypes;

			if ( try_next_increment != -1 )
			{
				increment_random_axis = try_next_increment;
				decrement_random_axis = try_next_decrement;
				try_next_increment = -1;
				try_next_decrement = -1;
			}

			if ( land_count_is_locked )
			{
				if ( increment_random_axis == 0 ) continue;
				if ( decrement_random_axis == 0 ) continue;
			}
			if ( increment_random_axis == decrement_random_axis ) continue;
			if ( current.deck.m_cards[ increment_random_axis ] == deck_size ) continue;
			if ( current.deck.m_cards[ decrement_random_axis ] == 0 ) continue;

			if ( tested_directions[ increment_random_axis ][ decrement_random_axis ] )
			{
				int untested_directions_count = 0;
				for ( int increment_direction = 0; increment_direction < Card::NumberOfCurveRelevantCardTypes; ++increment_direction )
				{
					for ( int decrement_direction = 0; decrement_direction < Card::NumberOfCurveRelevantCardTypes; ++decrement_direction )
					{
						if ( land_count_is_locked )
						{
							if ( increment_direction == 0 ) continue;
							if ( decrement_direction == 0 ) continue;
						}
						if ( increment_direction == decrement_direction ) continue;
						if ( current.deck.m_cards[ increment_direction ] == deck_size ) continue;
						if ( current.deck.m_cards[ decrement_direction ] == 0 ) continue;
						if ( !tested_directions[ increment_direction ][ decrement_direction ] )
						{
							++untested_directions_count;
						}
					}
				}

				if ( untested_directions_count == 0 ) break;

				continue;
			}

			++iteration;

			CDeck test;
			test.deck = current.deck;
			++test.deck.m_cards[ increment_random_axis ];
			--test.deck.m_cards[ decrement_random_axis ];

			int cache_size = cache.size();

			test.value = determine_metric_for_deck( test.deck, situation, mulligan_condition, metric, cache );
			const bool result = test.value >= current.value;

			if ( VERBOSE )
			{
				print_iteration( stderr, iteration, cache_size == cache.size(), test, result );
			}

			if ( result )
			{
				current = test;
				memset( tested_directions, 0, sizeof( tested_directions ) );

				// no need to retest where we came from
				tested_directions[ decrement_random_axis ][ increment_random_axis ] = true;

				try_next_increment = increment_random_axis;
				try_next_decrement = decrement_random_axis;
			}
			else
			{
				tested_directions[ increment_random_axis ][ decrement_random_axis ] = true;
			}
		}
	}

	if ( VERBOSE )
	{
		fprintf( stderr, "       Found local maximum: " );
		print_deck( stderr, current.deck );
		fprintf( stderr, ".\n" );
		fprintf( stderr, "       Hill climb complete.\n" );
		fflush( stderr );
	}

	return current;
}

template< typename TMulliganCondition, typename TMetric >
CDeck hill_climb_with_shortcuts( CPile initial_deck, const CSituation& situation, TMulliganCondition mulligan_condition, TMetric metric, CDeckCache& shortcut_cache, CDeckCache& standard_cache )
{
	assert( initial_deck.GetSize() == DECK_SIZE );

	CSituation local_situation = situation;

	local_situation.play_or_draw_scheme = situation.play_or_draw_scheme;
	if ( situation.play_or_draw_scheme == Averaged_UsePlayOrDrawShortcut )
	{
		if ( VERBOSE ) fprintf( stderr, "     Using play/draw shortcut. Initial maximization done on the play.\n" );
		local_situation.play_or_draw_scheme = OnThePlay;
	}

	CDeck current;
	if ( local_situation.use_mulligan_shortcut )
	{
		if ( VERBOSE ) fprintf( stderr, "     Using mulligan shortcut. Initial maximization done without mulligans.\n" );
		if ( VERBOSE ) fprintf( stderr, "     Starting initial, shortcut maximization.\n" );

		auto no_mull = []( TPlayOrDraw play_or_draw_scheme, const CProbabilityDistributionEntry& entry ) { return false; };
		current = hill_climb( initial_deck, local_situation, no_mull, false, metric, shortcut_cache );
	}
	else
	{
		current = hill_climb( initial_deck, local_situation, mulligan_condition, true, metric, shortcut_cache );
	}

	if ( situation.play_or_draw_scheme == Averaged_UsePlayOrDrawShortcut || situation.use_mulligan_shortcut )
	{
		if ( situation.play_or_draw_scheme == Averaged_UsePlayOrDrawShortcut )
		{
			if ( VERBOSE ) fprintf( stderr, "     Removing play/draw shortcut.\n" );
			local_situation.play_or_draw_scheme = Averaged;
		}

		if ( situation.use_mulligan_shortcut )
		{
			if ( VERBOSE ) fprintf( stderr, "     Removing mulligan shortcut.\n" );
		}
		if ( VERBOSE ) fprintf( stderr, "     Starting final maximization.\n" );

		current = hill_climb( current.deck, local_situation, mulligan_condition, true, metric, standard_cache );
	}

	if ( VERBOSE ) fflush( stderr );

	return current;
}

template< typename TMulliganCondition, typename TMetric >
CDeck find_deck_that_maximizes_metric( const CSituation& situation, TMulliganCondition mulligan_condition, TMetric metric )
{
	if ( VERBOSE )
	{
		fprintf( stderr, "   Finding deck that maximizes metric.\n" );
		fprintf( stderr, "   VERBOSE on. Verbose output sent to stderr.\n" );
		fprintf( stderr, "   Running hill climb multiple times with various starting points.\n" );
		fprintf( stderr, "   Finding deck that maximizes metric. Working...\n" );
	}

	const bool land_count_is_locked = situation.locked_land_count != -1;

	CDeck results[ 64 ];
	int next_result = 0;
	CPile deck;

	CDeckCache shortcut_cache;
	CDeckCache standard_cache;

	{
		deck.m_cards[ Card::Land ] = land_count_is_locked ? situation.locked_land_count : LAND_COUNT;
		deck.m_cards[ Card::DeadCard ] = situation.number_of_dead_cards;

		for ( int mana_cost = Card::ManaCost_1; mana_cost < Card::HighestManaCost; ++mana_cost )
		{
			deck.m_cards[ mana_cost ] = ( DECK_SIZE - deck.m_cards[ Card::Land ] - deck.m_cards[ Card::DeadCard ] ) / ( Card::HighestManaCost );
		}
		deck.m_cards[ Card::HighestManaCost ] = 0;

		deck.m_cards[ Card::HighestManaCost ] = DECK_SIZE - deck.GetSize();

		if ( VERBOSE )
		{
			fprintf( stderr, "   Running using even-distribution starting point: " );
			print_deck( stderr, deck );
			fprintf( stderr, ".\n" );
		}

		results[ next_result++ ] = hill_climb_with_shortcuts( deck, situation, mulligan_condition, metric, shortcut_cache, standard_cache );
	}

	const int first_card_type = land_count_is_locked ? Card::ManaCost_1 : Card::Land;
	for ( int card_type = first_card_type; card_type < Card::NumberOfCurveRelevantCardTypes; ++card_type )
	{
		deck.m_cards[ Card::Land ] = land_count_is_locked ? situation.locked_land_count : 20;
		deck.m_cards[ Card::DeadCard ] = situation.number_of_dead_cards;

		for ( int mana_cost = Card::ManaCost_1; mana_cost < Card::NumberOfCurveRelevantCardTypes; ++mana_cost )
		{
			deck.m_cards[ mana_cost ] = 0;
		}

		deck.m_cards[ card_type ] += DECK_SIZE - deck.m_cards[ Card::Land ] - deck.m_cards[ Card::DeadCard ];

		if ( VERBOSE )
		{
			fprintf( stderr, "   Running using single-mana-cost (%i) starting point: ", card_type );
			print_deck( stderr, deck );
			fprintf( stderr, ".\n" );
		}

		results[ next_result++ ] = hill_climb_with_shortcuts( deck, situation, mulligan_condition, metric, shortcut_cache, standard_cache );
	}

	const int NUM_RANDOM_TESTS = 5;
	for ( int count = 0; count < NUM_RANDOM_TESTS; ++count )
	{
		deck.m_cards[ Card::Land ] = land_count_is_locked ? situation.locked_land_count : 0;
		deck.m_cards[ Card::DeadCard ] = situation.number_of_dead_cards;

		for ( int mana_cost = Card::ManaCost_1; mana_cost < Card::NumberOfCurveRelevantCardTypes; ++mana_cost )
		{
			deck.m_cards[ mana_cost ] = 0;
		}

		for ( int card_count = deck.m_cards[ Card::Land ] + deck.m_cards[ Card::DeadCard ]; card_count < DECK_SIZE; )
		{
			int axis = random() % Card::NumberOfCurveRelevantCardTypes;
			if ( !land_count_is_locked || axis != Card::Land )
			{
				deck.m_cards[ axis ]++;
				 ++card_count;
			}
		}

		if ( VERBOSE )
		{
			fprintf( stderr, "   Running using random starting point: " );
			print_deck( stderr, deck );
			fprintf( stderr, ".\n" );
		}

		results[ next_result++ ] = hill_climb_with_shortcuts( deck, situation, mulligan_condition, metric, shortcut_cache, standard_cache );
	}

	if ( VERBOSE )
	{
		fprintf( stderr, "   Hill climb completed for all start points.\n" );
		fprintf( stderr, "   Found these local maxima:\n" );
	}

	CDeck result = results[ 0 ];
	if ( VERBOSE )
	{
		fprintf( stderr, "     " );
		print_deck( stderr, results[ 0 ].deck );
		fprintf( stderr, " m = %.4f\n", results[ 0 ].value );
	}
	for ( int result_index = 1; result_index < next_result; ++result_index )
	{
		if ( VERBOSE )
		{
			fprintf( stderr, "     " );
			print_deck( stderr, results[ result_index ].deck );
			fprintf( stderr, " m = %.4f\n", results[ result_index ].value );
		}

		if ( results[ result_index ].value > result.value )
		{
			result = results[ result_index ];
		}
	}

	if ( VERBOSE )
	{
		fprintf( stderr, "   Found global maximum: " );
		print_deck( stderr, result.deck );
		fprintf( stderr, ".\n" );
		fprintf( stderr, "   Maximization complete.\n" );
		fflush( stderr );
	}

	return result;
}

void part_01()
{
	auto mulligan_rules = []( TPlayOrDraw play_or_draw, const CProbabilityDistributionEntry& entry ) -> bool
	{
		const int lands = entry.m_player.m_hand.m_cards[ Card::Land ];

		if ( play_or_draw == Play )
		{
			if ( entry.m_player.m_hand.GetSize() == 7 )
			{
				return lands < 3 || lands > 5;
			}
			if ( entry.m_player.m_hand.GetSize() == 6 )
			{
				return lands < 2 || lands > 4;
			}
			if ( entry.m_player.m_hand.GetSize() == 5 )
			{
				return lands < 1 || lands > 4;
			}
			return false;
		}
		else
		{
			if ( entry.m_player.m_hand.GetSize() == 7 )
			{
				return lands < 3 || lands > 5;
			}
			if ( entry.m_player.m_hand.GetSize() == 6 )
			{
				return lands < 2 || lands > 5;
			}
			if ( entry.m_player.m_hand.GetSize() == 5 )
			{
				return lands < 2 || lands > 4;
			}
			return false;
		}
	};

	auto metric_total_mana = []( int turn, CProbabilityDistributionEntry& entry )
	{
		int mana = std::min( turn, entry.m_player.m_hand.m_cards[ Card::Land ] );
		const int highest_possible_mana_cost = std::min( mana, int( Card::HighestManaCost ) );

		for ( int mana_cost = highest_possible_mana_cost; mana_cost >= Card::ManaCost_1; )
		{
			if ( mana == 0 ) break;
			if ( mana >= mana_cost && entry.m_player.m_hand.m_cards[ mana_cost ] > 0 )
			{
				--entry.m_player.m_hand.m_cards[ mana_cost ];
				mana -= mana_cost;
				entry.m_user_data += mana_cost;
			}
			else
			{
				--mana_cost;
			}
		}
	};

	auto metric_perfect_curve = []( int turn, CProbabilityDistributionEntry& entry )
	{
		// states which are perfect have user-data 0.0; states which are not have user-data -1.0
		// maximizing metric results in highest probability of perfection

		if ( entry.m_user_data == -1.0 ) return;

		int mana_spent = 0;
		int mana = std::min( turn, entry.m_player.m_hand.m_cards[ Card::Land ] );
		const int highest_possible_mana_cost = std::min( mana, int( Card::HighestManaCost ) );

		for ( int mana_cost = highest_possible_mana_cost; mana_cost >= Card::ManaCost_1; )
		{
			if ( mana == 0 ) break;
			if ( mana >= mana_cost && entry.m_player.m_hand.m_cards[ mana_cost ] > 0 )
			{
				--entry.m_player.m_hand.m_cards[ mana_cost ];
				mana -= mana_cost;
				mana_spent += mana_cost;
			}
			else
			{
				--mana_cost;
			}
		}

		if ( mana_spent < turn )
		{
			entry.m_user_data = -1.0;
		}
	};

	auto metric_perfect_curve_no_hole_filling = []( int turn, CProbabilityDistributionEntry& entry )
	{
		// states which are perfect have user-data 0.0; states which are not have user-data -1.0
		// maximizing metric results in highest probability of perfection

		if ( entry.m_user_data == -1.0 ) return;

		int mana_spent = 0;
		int mana = std::min( turn, entry.m_player.m_hand.m_cards[ Card::Land ] );
		const int highest_possible_mana_cost = std::min( mana, int( Card::HighestManaCost ) );

		for ( int mana_cost = highest_possible_mana_cost; mana_cost >= Card::ManaCost_1; )
		{
			if ( mana == 0 ) break;
			if ( mana >= mana_cost && entry.m_player.m_hand.m_cards[ mana_cost ] > 0 )
			{
				--entry.m_player.m_hand.m_cards[ mana_cost ];
				mana -= mana_cost;
				mana_spent += mana_cost;
				break;
			}
			else
			{
				--mana_cost;
			}
		}

		if ( mana_spent < turn )
		{
			entry.m_user_data = -1.0;
		}
	};

	auto metric_spell_count = []( int turn, CProbabilityDistributionEntry& entry )
	{
		int mana = std::min( turn, entry.m_player.m_hand.m_cards[ Card::Land ] );
		const int highest_possible_mana_cost = std::min( mana, int( Card::HighestManaCost ) );

		for ( int mana_cost = highest_possible_mana_cost; mana_cost >= Card::ManaCost_1; )
		{
			if ( mana == 0 ) break;
			if ( mana >= mana_cost && entry.m_player.m_hand.m_cards[ mana_cost ] > 0 )
			{
				--entry.m_player.m_hand.m_cards[ mana_cost ];
				mana -= mana_cost;
				entry.m_user_data += 1.0;
			}
			else
			{
				--mana_cost;
			}
		}
	};

	auto metric_first_spell_total_mana = []( int turn, CProbabilityDistributionEntry& entry )
	{
		int mana = std::min( turn, entry.m_player.m_hand.m_cards[ Card::Land ] );
		const int highest_possible_mana_cost = std::min( mana, int( Card::HighestManaCost ) );

		for ( int mana_cost = highest_possible_mana_cost; mana_cost >= Card::ManaCost_1; )
		{
			if ( mana == 0 ) break;
			if ( mana >= mana_cost && entry.m_player.m_hand.m_cards[ mana_cost ] > 0 )
			{
				--entry.m_player.m_hand.m_cards[ mana_cost ];
				mana -= mana_cost;
				entry.m_user_data += mana_cost;
				break;
			}
			else
			{
				--mana_cost;
			}
		}
	};

	auto metric_first_spell_no_break_total_mana = []( int turn, CProbabilityDistributionEntry& entry )
	{
		bool spell_cast = false;

		int mana = std::min( turn, entry.m_player.m_hand.m_cards[ Card::Land ] );
		const int highest_possible_mana_cost = std::min( mana, int( Card::HighestManaCost ) );

		for ( int mana_cost = highest_possible_mana_cost; mana_cost >= Card::ManaCost_1; )
		{
			if ( mana == 0 ) break;
			if ( mana >= mana_cost && entry.m_player.m_hand.m_cards[ mana_cost ] > 0 )
			{
				--entry.m_player.m_hand.m_cards[ mana_cost ];
				mana -= mana_cost;
				if ( !spell_cast )
				{
					entry.m_user_data += mana_cost;
				}
				spell_cast = true;
			}
			else
			{
				--mana_cost;
			}
		}
	};

	auto metric_adjusted_mana = []( int turn, CProbabilityDistributionEntry& entry )
	{
		int mana = std::min( turn, entry.m_player.m_hand.m_cards[ Card::Land ] );
		const int highest_possible_mana_cost = std::min( mana, int( Card::HighestManaCost ) );

		for ( int mana_cost = highest_possible_mana_cost; mana_cost >= Card::ManaCost_1; )
		{
			if ( mana == 0 ) break;
			if ( mana >= mana_cost && entry.m_player.m_hand.m_cards[ mana_cost ] > 0 )
			{
				--entry.m_player.m_hand.m_cards[ mana_cost ];
				mana -= mana_cost;
				entry.m_user_data += mana_cost * mana_cost / double( turn );
			}
			else
			{
				--mana_cost;
			}
		}
	};

	auto metric_damage = []( int turn, CProbabilityDistributionEntry& entry )
	{
		int mana = std::min( turn, entry.m_player.m_hand.m_cards[ Card::Land ] );
		const int highest_possible_mana_cost = std::min( mana, int( Card::HighestManaCost ) );

		for ( int mana_cost = highest_possible_mana_cost; mana_cost >= std::max( 1, turn - 1 ); )
		{
			if ( mana == 0 ) break;
			if ( mana >= mana_cost && entry.m_player.m_hand.m_cards[ mana_cost ] > 0 )
			{
				--entry.m_player.m_hand.m_cards[ mana_cost ];
				mana -= mana_cost;

				const double power_mapping[] = { -1.0, 1.0, 2.0, 2.5, 3.0, 4.0 };
				const double power = power_mapping[ mana_cost ];

				if ( mana_cost == turn )
				{
					entry.m_user_data += 2 * power;
				}
				else if ( mana_cost == turn - 1 )
				{
					entry.m_user_data += power;
				}
				else
				{
					entry.m_user_data += 0.0;
				}
			}
			else
			{
				--mana_cost;
			}
		}
	};

	{
		CTable table( "Average mana used and probability of perfect mana use in the first 5 turns for decks with various mana curves." );
		table.AddColumn( "land" );
		table.AddColumn( "CC1" );
		table.AddColumn( "CC2" );
		table.AddColumn( "CC3" );
		table.AddColumn( "CC4" );
		table.AddColumn( "CC5" );
		table.AddColumn( "perfect mana use" );
		table.AddColumn( "avg mana used" );

		CSituation situation;
		situation.locked_land_count = 17;
		situation.number_of_turns = 5;
		situation.play_or_draw_scheme = Averaged;
		situation.use_mulligan_shortcut = false;

		int decks[][ Card::NumberOfCardTypes ] =
		{
			{ 17, 3, 5, 6, 5, 4, 0 },
			{ 17, 2, 6, 6, 6, 3, 0 },
			{ 17, 1, 9, 7, 4, 2, 0 },
			{ 17, 2, 8, 7, 4, 2, 0 },
			{ 17, 0, 10, 7, 4, 2, 0 },
			{ 17, 1, 9, 6, 4, 3, 0 },
			{ 17, 1, 9, 6, 5, 2, 0 },
			{ 17, 1, 8, 7, 5, 2, 0 }
		};

		for ( int deck_idx = 0; deck_idx < std::extent< decltype( decks ) >::value; ++deck_idx )
		{
			CPile deck;

			deck.m_cards[ Card::Land ] = decks[ deck_idx ][ Card::Land ];
			deck.m_cards[ Card::ManaCost_1 ] = decks[ deck_idx ][ Card::ManaCost_1 ];
			deck.m_cards[ Card::ManaCost_2 ] = decks[ deck_idx ][ Card::ManaCost_2 ];
			deck.m_cards[ Card::ManaCost_3 ] = decks[ deck_idx ][ Card::ManaCost_3 ];
			deck.m_cards[ Card::ManaCost_4 ] = decks[ deck_idx ][ Card::ManaCost_4 ];
			deck.m_cards[ Card::ManaCost_5 ] = decks[ deck_idx ][ Card::ManaCost_5 ];
			deck.m_cards[ Card::DeadCard ] = decks[ deck_idx ][ Card::DeadCard ];
			assert( deck.GetSize() == 40 );

			table.AddInteger( deck.m_cards[ Card::Land ] );
			table.AddInteger( deck.m_cards[ Card::ManaCost_1 ] );
			table.AddInteger( deck.m_cards[ Card::ManaCost_2 ] );
			table.AddInteger( deck.m_cards[ Card::ManaCost_3 ] );
			table.AddInteger( deck.m_cards[ Card::ManaCost_4 ] );
			table.AddInteger( deck.m_cards[ Card::ManaCost_5 ] );

			const double probability_of_perfect_mana = calculate_metric_for_deck( deck, situation, mulligan_rules, metric_perfect_curve );
			table.AddProbability( 1.0 + probability_of_perfect_mana );

			const double average_mana_spent = calculate_metric_for_deck( deck, situation, mulligan_rules, metric_total_mana );
			table.AddNumber( average_mana_spent, 2 );
		}

		table.Print();
	}

	{
		if ( VERBOSE )
		{
			fprintf( stderr, "\n" );
			fprintf( stderr, " Testing metric: total mana spent.\n" );
		}

		CSituation situation;
		situation.locked_land_count = LAND_COUNT;
		situation.number_of_turns = 5;
		situation.number_of_dead_cards = 0;
		situation.play_or_draw_scheme = Averaged_UsePlayOrDrawShortcut;
		situation.use_mulligan_shortcut = true;

		CDeck deck_that_maximizes_metric = find_deck_that_maximizes_metric( situation, mulligan_rules, metric_total_mana );

		print_deck_table(
			"What mana curve results in maximum average mana used in the first 5 turns?", 0,
			deck_that_maximizes_metric,
			"average mana used"
		);
	}

	{
		if ( VERBOSE )
		{
			fprintf( stderr, "\n" );
			fprintf( stderr, " Testing metric: perfect mana usage.\n" );
		}

		CSituation situation;
		situation.locked_land_count = LAND_COUNT;
		situation.number_of_turns = 5;
		situation.number_of_dead_cards = 0;
		situation.play_or_draw_scheme = Averaged_UsePlayOrDrawShortcut;
		situation.use_mulligan_shortcut = true;

		CDeck deck_that_maximizes_metric = find_deck_that_maximizes_metric( situation, mulligan_rules, metric_perfect_curve );

		situation.play_or_draw_scheme = Averaged;
		const double average_mana_spent = calculate_metric_for_deck( deck_that_maximizes_metric.deck, situation, mulligan_rules, metric_total_mana );

		print_deck_table(
			"What mana curve results in the maximum probability of perfect mana use in the first 5 turns?", 0,
			deck_that_maximizes_metric,
			"probability of perfect mana usage",
			"average mana used",
			average_mana_spent
		);
	}

	{
		if ( VERBOSE )
		{
			fprintf( stderr, "\n" );
			fprintf( stderr, " Testing metric: perfect mana usage without hole filling.\n" );
		}

		CSituation situation;
		situation.locked_land_count = LAND_COUNT;
		situation.number_of_turns = 5;
		situation.number_of_dead_cards = 0;
		situation.play_or_draw_scheme = Averaged_UsePlayOrDrawShortcut;
		situation.use_mulligan_shortcut = true;

		CDeck deck_that_maximizes_metric = find_deck_that_maximizes_metric( situation, mulligan_rules, metric_perfect_curve_no_hole_filling );

		situation.play_or_draw_scheme = Averaged;
		const double average_mana_spent = calculate_metric_for_deck( deck_that_maximizes_metric.deck, situation, mulligan_rules, metric_total_mana );

		print_deck_table(
			"What mana curve results in the maximum probability of casting the costliest possible spell every turn for the first 5 turns?", 0,
			deck_that_maximizes_metric,
			"probability of absolutely perfect mana usage",
			"average mana used",
			average_mana_spent
		);
	}

	{
		if ( VERBOSE )
		{
			fprintf( stderr, "\n" );
			fprintf( stderr, " Testing metric: total mana spent first spell.\n" );
		}

		CSituation situation;
		situation.locked_land_count = LAND_COUNT;
		situation.number_of_turns = 5;
		situation.number_of_dead_cards = 0;
		situation.play_or_draw_scheme = Averaged_UsePlayOrDrawShortcut;
		situation.use_mulligan_shortcut = true;

		CDeck deck_that_maximizes_metric = find_deck_that_maximizes_metric( situation, mulligan_rules, metric_first_spell_total_mana );

		print_deck_table(
			"What mana curve results in maximum average mana used in the first 5 turns, counting only mana spent on the first spell each turn?", 0,
			deck_that_maximizes_metric
		);
	}

	{
		CTable table( "Credit given by the balanced metric for casting a spell on the given turn." );
		table.AddColumn( "casting cost" );
		table.AddColumn( "  1" );
		table.AddColumn( "  2" );
		table.AddColumn( "  3" );
		table.AddColumn( "  4" );
		table.AddColumn( "  5" );
		for ( int casting_cost = 1; casting_cost <= 5; ++casting_cost )
		{
			table.AddInteger( casting_cost );
			for ( int turn = 1; turn <= 5; ++turn )
			{
				if ( turn >= casting_cost )
				{
					table.AddNumber( casting_cost * casting_cost / double( turn ), 2 );
				}
				else
				{
					table.AddString( " -" );
				}
			}
		}
		table.Print();
	}

	{
		if ( VERBOSE )
		{
			fprintf( stderr, "\n" );
			fprintf( stderr, " Testing metric: balanced.\n" );
		}

		CSituation situation;
		situation.locked_land_count = LAND_COUNT;
		situation.number_of_turns = 5;
		situation.number_of_dead_cards = 0;
		situation.play_or_draw_scheme = Averaged_UsePlayOrDrawShortcut;
		situation.use_mulligan_shortcut = true;

		CDeck deck_that_maximizes_metric = find_deck_that_maximizes_metric( situation, mulligan_rules, metric_adjusted_mana );

		print_deck_table(
			"What mana curve maximizes a balanced metric?", 0,
			deck_that_maximizes_metric
		);
	}

	{
		CTable table( "Maximized mana curve for decks with various numbers of off-curve spells." );
		table.AddColumn( "off-curve spells" );
		table.AddColumn( "land" );
		table.AddColumn( "CC1" );
		table.AddColumn( "CC2" );
		table.AddColumn( "CC3" );
		table.AddColumn( "CC4" );
		table.AddColumn( "CC5" );

		for ( int number_of_dead_cards = 1; number_of_dead_cards <= 6; ++number_of_dead_cards )
		{
			if ( VERBOSE )
			{
				fprintf( stderr, "\n" );
				fprintf( stderr, " Testing metric: balanced.\n" );
			}

			CSituation situation;
			situation.locked_land_count = LAND_COUNT;
			situation.number_of_turns = 5;
			situation.number_of_dead_cards = number_of_dead_cards;
			situation.play_or_draw_scheme = Averaged_UsePlayOrDrawShortcut;
			situation.use_mulligan_shortcut = true;

			CDeck deck_that_maximizes_metric = find_deck_that_maximizes_metric( situation, mulligan_rules, metric_adjusted_mana );

			print_deck_table(
				"What mana curve maximizes a balanced metric with %d off-curve spell(s) in the deck?", number_of_dead_cards,
				deck_that_maximizes_metric
			);

			table.AddInteger( deck_that_maximizes_metric.deck.m_cards[ Card::DeadCard ] );
			table.AddInteger( deck_that_maximizes_metric.deck.m_cards[ Card::Land ] );
			table.AddInteger( deck_that_maximizes_metric.deck.m_cards[ Card::ManaCost_1 ] );
			table.AddInteger( deck_that_maximizes_metric.deck.m_cards[ Card::ManaCost_2 ] );
			table.AddInteger( deck_that_maximizes_metric.deck.m_cards[ Card::ManaCost_3 ] );
			table.AddInteger( deck_that_maximizes_metric.deck.m_cards[ Card::ManaCost_4 ] );
			table.AddInteger( deck_that_maximizes_metric.deck.m_cards[ Card::ManaCost_5 ] );
		}

		table.Print();
	}

	{
		if ( VERBOSE )
		{
			fprintf( stderr, "\n" );
			fprintf( stderr, " Testing metric: total mana spent.\n" );
		}

		CSituation situation;
		situation.locked_land_count = -1;
		situation.number_of_turns = 5;
		situation.number_of_dead_cards = 0;
		situation.play_or_draw_scheme = Averaged_UsePlayOrDrawShortcut;
		situation.use_mulligan_shortcut = true;

		CDeck deck_that_maximizes_metric = find_deck_that_maximizes_metric( situation, mulligan_rules, metric_total_mana );

		print_deck_table(
			"What mana curve results in maximum average mana used in the first 5 turns when the deck's land count is allowed to vary?", 0,
			deck_that_maximizes_metric,
			"average mana used"
		);
	}

	{
		if ( VERBOSE )
		{
			fprintf( stderr, "\n" );
			fprintf( stderr, " Testing metric: balanced.\n" );
		}

		CSituation situation;
		situation.locked_land_count = -1;
		situation.number_of_turns = 5;
		situation.number_of_dead_cards = 0;
		situation.play_or_draw_scheme = Averaged_UsePlayOrDrawShortcut;
		situation.use_mulligan_shortcut = true;

		CDeck deck_that_maximizes_metric = find_deck_that_maximizes_metric( situation, mulligan_rules, metric_adjusted_mana );

		print_deck_table(
			"What mana curve maximizes a balanced metric when the deck's land count is allowed to vary?", 0,
			deck_that_maximizes_metric
		);
	}
}

};

void article_15()
{
	A15_P01::part_01();
}