#ifndef MAGIC_PROBABILITY_TOOLKIT_H
#define MAGIC_PROBABILITY_TOOLKIT_H

#include <cassert>
#include <functional>
#include <list>
#include <type_traits>

namespace mpt
{

enum TPlayOrDraw
{
	Play,
	Draw
};

enum
{
	HandSizePlusOne = 7 + 1
};

template< class TCardSet >
struct CPileTemplate
{
	typedef TCardSet TCardSetType;

	union
	{
		int m_cards[ TCardSet::NumberOfCardTypes ];
		struct
		{
			int m_lands;
			int m_spells;
		};
	};

	bool operator==( const CPileTemplate& pile ) const
	{
		for ( int card_type = 0; card_type < TCardSet::NumberOfCardTypes; ++card_type )
		{
			if ( m_cards[ card_type ] != pile.m_cards[ card_type ] ) return false;
		}
		return true;
	}

	int GetSize() const
	{
		int total = 0;
		for ( int card_type = 0; card_type < TCardSet::NumberOfCardTypes; ++card_type )
		{
			total += m_cards[ card_type ];
		}
		return total;
	}

	double GetCardTypeRatio( int card_type ) const
	{
		assert( card_type < TCardSet::NumberOfCardTypes );
		return double( m_cards[ card_type ] ) / GetSize();
	}

	double GetLandRatio() const { return GetCardTypeRatio( CardSet::IgnoringColor::Land ); }
	double GetSpellRatio() const { return GetCardTypeRatio( CardSet::IgnoringColor::Spell ); }
};

namespace CardSet
{
	struct IgnoringColor
	{
		enum
		{
			Land,
			Spell,
			NumberOfCardTypes
		};

		enum { NumberOfHandsPerHandSize = HandSizePlusOne };
		typedef CPileTemplate< IgnoringColor > TPileType;

		static bool sGetHandAndLibrary( TPileType& hand, TPileType& library, const TPileType& deck, int hand_size, int hand_index )
		{
			hand.m_cards[ Land ] = hand_index;
			hand.m_cards[ Spell ] = hand_size - hand_index;

			for ( int card_type = 0; card_type < NumberOfCardTypes; ++card_type )
			{
				library.m_cards[ card_type ] = deck.m_cards[ card_type ] - hand.m_cards[ card_type ];
				if ( library.m_cards[ card_type ] < 0 ) return false;
			}
			return true;
		}

		static bool sIsBeyondLastHand( int hand_size, int hand_index )
		{
			return hand_index > hand_size;
		}

		static int sGetIndex( TPileType hand )
		{
			return hand.m_cards[ Land ];
		}
	};

	struct IgnoringColorWithRamp
	{
		enum
		{
			Land,
			Ramp,
			Spell,
			RampInPlay,
			Turn,
			NumberOfCardTypes
		};

		enum { NumberOfHandsPerHandSize = HandSizePlusOne * HandSizePlusOne };
		typedef CPileTemplate< IgnoringColorWithRamp > TPileType;

		static bool sGetHandAndLibrary( TPileType& hand, TPileType& library, const TPileType& deck, int hand_size, int hand_index )
		{
			hand.m_cards[ Land ] = hand_index / HandSizePlusOne;
			hand.m_cards[ Ramp ] = hand_index % HandSizePlusOne;
			hand.m_cards[ RampInPlay ] = 0;

			hand.m_cards[ Spell ] = hand_size - hand.m_cards[ Land ] - hand.m_cards[ Ramp ];
			if ( hand.m_cards[ Spell ] < 0 ) return false;

			hand.m_cards[ Turn ] = 0;

			for ( int card_type = 0; card_type < NumberOfCardTypes; ++card_type )
			{
				library.m_cards[ card_type ] = deck.m_cards[ card_type ] - hand.m_cards[ card_type ];
				if ( library.m_cards[ card_type ] < 0 ) return false;
			}
			return true;
		}

		static bool sIsBeyondLastHand( int hand_size, int hand_index )
		{
			return hand_index >= NumberOfHandsPerHandSize;
		}

		static int sGetIndex( TPileType hand )
		{
			return hand.m_cards[ Land ] * HandSizePlusOne + hand.m_cards[ Ramp ];
		}
	};

	struct IgnoringColorWithCantrips
	{
		enum
		{
			Land,
			Spell,
			Cantrip,
			Turn,
			CardsToDraw,
			ManaSpent,
			NumberOfCardTypes,
		};

		enum { NumberOfHandsPerHandSize = HandSizePlusOne * HandSizePlusOne };
		typedef CPileTemplate< IgnoringColorWithCantrips > TPileType;

		static bool sGetHandAndLibrary( TPileType& hand, TPileType& library, const TPileType& deck, int hand_size, int hand_index )
		{
			hand.m_cards[ Land ] = hand_index / HandSizePlusOne;
			hand.m_cards[ Cantrip ] = hand_index % HandSizePlusOne;

			hand.m_cards[ Spell ] = hand_size - hand.m_cards[ Land ] - hand.m_cards[ Cantrip ];
			if ( hand.m_cards[ Spell ] < 0 ) return false;

			hand.m_cards[ Turn ] = 0;
			hand.m_cards[ CardsToDraw ] = 0;
			hand.m_cards[ ManaSpent ] = 0;

			for ( int card_type = 0; card_type < NumberOfCardTypes; ++card_type )
			{
				library.m_cards[ card_type ] = deck.m_cards[ card_type ] - hand.m_cards[ card_type ];
				if ( library.m_cards[ card_type ] < 0 ) return false;
			}
			return true;
		}

		static bool sIsBeyondLastHand( int hand_size, int hand_index )
		{
			return hand_index >= NumberOfHandsPerHandSize;
		}

		static int sGetIndex( TPileType hand )
		{
			return hand.m_cards[ Land ] * HandSizePlusOne + hand.m_cards[ Cantrip ];
		}
	};

	struct TwoLandColors
	{
		enum
		{
			LandC,
			LandD,
			Spell,
			Turn,
			NumberOfCardTypes
		};

		enum { NumberOfHandsPerHandSize = HandSizePlusOne * HandSizePlusOne };
		typedef CPileTemplate< TwoLandColors > TPileType;

		static bool sGetHandAndLibrary( TPileType& hand, TPileType& library, const TPileType& deck, int hand_size, int hand_index )
		{
			hand.m_cards[ LandC ] = hand_index / HandSizePlusOne;
			hand.m_cards[ LandD ] = hand_index % HandSizePlusOne;

			hand.m_cards[ Spell ] = hand_size - hand.m_cards[ LandC ] - hand.m_cards[ LandD ];
			if ( hand.m_cards[ Spell ] < 0 ) return false;

			hand.m_cards[ Turn ] = 0;

			for ( int card_type = 0; card_type < NumberOfCardTypes; ++card_type )
			{
				library.m_cards[ card_type ] = deck.m_cards[ card_type ] - hand.m_cards[ card_type ];
				if ( library.m_cards[ card_type ] < 0 ) return false;
			}
			return true;
		}

		static bool sIsBeyondLastHand( int hand_size, int hand_index )
		{
			return hand_index >= NumberOfHandsPerHandSize;
		}

		static int sGetIndex( TPileType hand )
		{
			return hand.m_cards[ LandC ] * HandSizePlusOne + hand.m_cards[ LandD ];
		}
	};

	struct TwoLandColorsWithFixing
	{
		enum
		{
			LandC,
			LandD,
			Fixing,
			Spell,
			Turn,
			NumberOfCardTypes
		};

		enum { NumberOfHandsPerHandSize = HandSizePlusOne * HandSizePlusOne * HandSizePlusOne };
		typedef CPileTemplate< TwoLandColorsWithFixing > TPileType;

		static bool sGetHandAndLibrary( TPileType& hand, TPileType& library, const TPileType& deck, int hand_size, int hand_index )
		{
			hand.m_cards[ LandC ] = hand_index / ( HandSizePlusOne * HandSizePlusOne );
			hand.m_cards[ LandD ] = ( hand_index / HandSizePlusOne ) % HandSizePlusOne;
			hand.m_cards[ Fixing ] = hand_index % HandSizePlusOne;

			const int land = hand.m_cards[ LandC ] + hand.m_cards[ LandD ] + hand.m_cards[ Fixing ];

			hand.m_cards[ Spell ] = hand_size - land;
			if ( hand.m_cards[ Spell ] < 0 ) return false;

			hand.m_cards[ Turn ] = 0;

			for ( int card_type = 0; card_type < NumberOfCardTypes; ++card_type )
			{
				library.m_cards[ card_type ] = deck.m_cards[ card_type ] - hand.m_cards[ card_type ];
				if ( library.m_cards[ card_type ] < 0 ) return false;
			}
			return true;
		}

		static bool sIsBeyondLastHand( int hand_size, int hand_index )
		{
			return hand_index >= NumberOfHandsPerHandSize;
		}

		static int sGetIndex( TPileType hand )
		{
			return hand.m_cards[ LandC ] * ( HandSizePlusOne * HandSizePlusOne ) + hand.m_cards[ LandD ] * HandSizePlusOne + hand.m_cards[ Fixing ];
		}
	};

	struct ManaCost
	{
		// no mulligan rule support
		enum
		{
			Land,
			ManaCost_1,
			ManaCost_2,
			ManaCost_3,
			ManaCost_4,
			ManaCost_5,
			DeadCard,
			NumberOfCardTypes,
			NumberOfCurveRelevantCardTypes = DeadCard,
			HighestManaCost = ManaCost_5
		};
	};

	struct IgnoringColorWithFixing
	{
		// no mulligan rule support
		enum
		{
			Land,
			Fixing,
			Spell,
			NumberOfCardTypes
		};
	};

	struct NumbersInAHat
	{
		// no mulligan rule support
		enum
		{
			Number0,
			Number1,
			Number2,
			Number3,
			Number4,
			Number5,
			Number6,
			Number7,
			Number8,
			Number9,
			Number10,
			Number11,
			Number12,
			Number13,
			Number14,
			Number15,
			Number16,
			NumberOfCardTypes
		};
	};
}

template< class TCardSet >
struct CPlayerTemplate
{
	typedef CPileTemplate< TCardSet > TPileType;

	TPileType m_library;
	TPileType m_hand;

	bool operator==( const CPlayerTemplate& player ) const
	{
		return m_library == player.m_library && m_hand == player.m_hand;
	}

	int GetDeckSize() const
	{
		return m_library.GetSize() + m_hand.GetSize();
	}

	int GetDeckCount( int card_type ) const
	{
		return m_library.m_cards[ card_type ] + m_hand.m_cards[ card_type ];
	}

	int GetDeckLandCount() const { return GetDeckCount( TCardSet::Land );	}
	int GetDeckSpellCount() const { return GetDeckCount( TCardSet::Spell ); }
};

typedef CPlayerTemplate< CardSet::IgnoringColor > CPlayer;

template< class TPlayer >
struct CPlayerProbabilityDistributionEntryTemplate
{
	enum { NumberOfPlayers = 1 };
	typedef TPlayer TPlayerType;

	double m_probability;
	bool m_dirty;
	double m_user_data;
	TPlayer m_player;

	explicit CPlayerProbabilityDistributionEntryTemplate( const TPlayer& player )
	{
		m_player = player;
	}

	bool operator==( const CPlayerProbabilityDistributionEntryTemplate& entry ) const
	{
		return m_player == entry.m_player && m_dirty == entry.m_dirty && m_user_data == entry.m_user_data;
	}

	TPlayer& GetPlayer( int player_idx )
	{
		return m_player;
	}

	const TPlayer& GetPlayer( int player_idx ) const
	{
		return m_player;
	}
};
typedef CPlayerProbabilityDistributionEntryTemplate< CPlayer > CPlayerProbabilityDistributionEntry;
typedef CPlayerProbabilityDistributionEntry CProbabilityDistributionEntry;

struct CGameProbabilityDistributionEntry
{
	enum { NumberOfPlayers = 2 };
	typedef CPlayer TPlayerType;

	double m_probability;
	bool m_dirty;
	double m_user_data;
	CPlayer m_players[ NumberOfPlayers ];

	explicit CGameProbabilityDistributionEntry( const CPlayer& player )
	{
		m_players[ 0 ] = player;
		m_players[ 1 ] = player;
	}

	bool operator==( const CGameProbabilityDistributionEntry& entry ) const
	{
		return m_players[ 0 ] == entry.m_players[ 0 ] && m_players[ 1 ] == entry.m_players[ 1 ] && m_dirty == entry.m_dirty && m_user_data == entry.m_user_data;
	}

	CPlayer& GetPlayer( int player_idx )
	{
		return m_players[ player_idx ];
	}

	const CPlayer& GetPlayer( int player_idx ) const
	{
		return m_players[ player_idx ];
	}
};

template< typename TCardSetType >
struct CMulliganRules
{
	typedef TCardSetType TCardSetType;

	bool m_table[ HandSizePlusOne ][ TCardSetType::NumberOfHandsPerHandSize ];

	bool IsMulliganAdvised( int hand_size, int hand_index ) const
	{
		return m_table[ hand_size ][ hand_index ];
	}
};

template< typename TCardSetType >
struct CMulliganRulesSet
{
	CMulliganRules< TCardSetType > m_on_the_play;
	CMulliganRules< TCardSetType > m_on_the_draw;
};

struct CTakeExtendedMeasurementResults
{
	double average;
	double standard_deviation;
};

template< class TProbabilityDistributionEntry >
struct CProbabilityDistributionTemplate
{
	typedef typename TProbabilityDistributionEntry TProbabilityDistributionEntry;
	typedef typename TProbabilityDistributionEntry::TPlayerType TPlayerType;
	typedef typename TPlayerType::TPileType TPileType;
	typedef typename TPileType::TCardSetType TCardSetType;

	std::list< TProbabilityDistributionEntry > m_distribution;

	CProbabilityDistributionTemplate() {};

	explicit CProbabilityDistributionTemplate( const CProbabilityDistributionTemplate& state )
	{
		for ( auto i = state.m_distribution.begin(); i != state.m_distribution.end(); ++i )
		{
			m_distribution.push_back( *i );
		}
	};

	void Init( const TPlayerType& player )
	{
		TProbabilityDistributionEntry initial_entry( player );
		initial_entry.m_probability = 1.0;
		initial_entry.m_dirty = false;
		initial_entry.m_user_data = 0.0;

		m_distribution.push_front( initial_entry );
	}
	
	CProbabilityDistributionTemplate( int deck_size, int land_count, int cards_in_hand = 0, int lands_in_hand = 0 )
	{
		assert( land_count <= deck_size );
		assert( lands_in_hand <= cards_in_hand );
		assert( lands_in_hand <= land_count );
		assert( cards_in_hand <= deck_size );
		assert( cards_in_hand - lands_in_hand <= deck_size - land_count );

		TPlayerType player;
		player.m_hand.m_lands = lands_in_hand;
		player.m_hand.m_spells = cards_in_hand - lands_in_hand;
		player.m_library.m_lands = land_count - player.m_hand.m_lands;
		player.m_library.m_spells = deck_size - cards_in_hand - player.m_library.m_lands;

		Init( player );
	}

	CProbabilityDistributionTemplate( const TPileType& library )
	{
		TPlayerType player;
		for ( int card_type = 0; card_type < TCardSetType::NumberOfCardTypes; ++card_type )
		{
			player.m_hand.m_cards[ card_type ] = 0;
		}
		player.m_library = library;

		Init( player );
	}

	CProbabilityDistributionTemplate( const TPileType& library, const TPileType& hand )
	{
		TPlayerType player;
		player.m_hand = hand;
		player.m_library = library;

		Init( player );
	}

	CProbabilityDistributionTemplate( const TProbabilityDistributionEntry& entry, const TPileType& library, int player_to_reset )
	{
		TProbabilityDistributionEntry initial_entry( entry );

		TPlayerType& player = initial_entry.GetPlayer( player_to_reset );
		for ( int card_type = 0; card_type < TCardSetType::NumberOfCardTypes; ++card_type )
		{
			player.m_hand.m_cards[ card_type ] = 0;
		}
		player.m_library = library;

		initial_entry.m_probability = 1.0;
		initial_entry.m_dirty = false;
		initial_entry.m_user_data = 0.0;

		m_distribution.push_front( initial_entry );
	}

	void AddEntry( TProbabilityDistributionEntry& new_entry )
	{
		assert( !new_entry.m_dirty );

		for ( auto i = m_distribution.begin(); i != m_distribution.end(); ++i )
		{
			TProbabilityDistributionEntry& entry = *i;
			if ( entry == new_entry )
			{
				entry.m_probability += new_entry.m_probability;
				return;
			}
		}
		m_distribution.push_front( new_entry );
	}
	
	void DrawCardFromEntry( const TProbabilityDistributionEntry& entry, int player = 0 )
	{
		assert( player < TProbabilityDistributionEntry::NumberOfPlayers );
		assert( entry.GetPlayer( player ).m_library.GetSize() > 0 );

		for ( int card_type = 0; card_type < TCardSetType::NumberOfCardTypes; ++card_type )
		{
			if ( entry.GetPlayer( player ).m_library.m_cards[ card_type ] > 0 )
			{
				TProbabilityDistributionEntry new_entry = entry;
				new_entry.m_probability *= entry.GetPlayer( player ).m_library.GetCardTypeRatio( card_type );
				new_entry.m_dirty = false;
				new_entry.GetPlayer( player ).m_hand.m_cards[ card_type ] += 1;
				new_entry.GetPlayer( player ).m_library.m_cards[ card_type ] -= 1;
				AddEntry( new_entry );
			}
		}
	}

	void MulliganFromEntry( const TProbabilityDistributionEntry& mulligan_from_entry, int player = 0 )
	{
		assert( player < TProbabilityDistributionEntry::NumberOfPlayers );

		TPileType deck;
		for ( int card_type = 0; card_type < TCardSetType::NumberOfCardTypes; ++card_type )
		{
			deck.m_cards[ card_type ]
				= mulligan_from_entry.GetPlayer( player ).m_library.m_cards[ card_type ]
				+ mulligan_from_entry.GetPlayer( player ).m_hand.m_cards[ card_type ];
		}

		const int previous_hand_size = mulligan_from_entry.GetPlayer( player ).m_hand.GetSize();
		const int new_hand_size = previous_hand_size - 1;
		assert( new_hand_size >= 0 );

		CProbabilityDistributionTemplate< TProbabilityDistributionEntry > sub_distribution( mulligan_from_entry, deck, player );
		sub_distribution.DrawCards( new_hand_size, player );

		for ( auto i = sub_distribution.m_distribution.begin(); i != sub_distribution.m_distribution.end(); ++i )
		{
			const TProbabilityDistributionEntry& sub_distribution_entry = *i;

			TProbabilityDistributionEntry new_entry = sub_distribution_entry;
			new_entry.m_probability *= mulligan_from_entry.m_probability;
			new_entry.m_dirty = false;
			new_entry.m_user_data = mulligan_from_entry.m_user_data;
			AddEntry( new_entry );
		}
	}
	
	template< typename TEndCondition >
	void DrawUntil( TEndCondition end_condition, int player = 0 )
	{
		bool end_condition_met;
		do
		{
			end_condition_met = true;

			for ( auto i = m_distribution.begin(); i != m_distribution.end(); )
			{
				const TProbabilityDistributionEntry& entry = *i;
				if ( end_condition( entry ) )
				{
					++i;
					continue;
				}

				end_condition_met = false;
				DrawCardFromEntry( entry, player );
				i = m_distribution.erase( i );
			}
		}
		while ( !end_condition_met );
	}
	
	template< typename TEndCondition >
	void TakeTurnsUntil( TEndCondition end_condition, int player = 0 )
	{
		bool end_condition_met;
		do
		{
			end_condition_met = true;

			for ( auto i = m_distribution.begin(); i != m_distribution.end(); )
			{
				TProbabilityDistributionEntry& entry = *i;
				if ( end_condition( entry ) )
				{
					++i;
					continue;
				}

				end_condition_met = false;
				DrawCardFromEntry( entry, player );
				i = m_distribution.erase( i );
			}
		}
		while ( !end_condition_met );
	}
	
	template< typename TEndCondition >
	void DrawIf( TEndCondition if_condition, int player = 0 )
	{
		for ( auto i = m_distribution.begin(); i != m_distribution.end(); )
		{
			const TProbabilityDistributionEntry& entry = *i;
			if ( !if_condition( entry ) )
			{
				++i;
				continue;
			}

			DrawCardFromEntry( entry, player );
			i = m_distribution.erase( i );
		}
	}

	template< typename TEndCondition >
	void AlternatingDrawUntil( TEndCondition end_condition )
	{
		int player = 1;

		bool end_condition_met;
		do
		{
			end_condition_met = true;

			for ( auto i = m_distribution.begin(); i != m_distribution.end(); )
			{
				const TProbabilityDistributionEntry& entry = *i;
				if ( end_condition( entry ) )
				{
					++i;
					continue;
				}

				end_condition_met = false;
				DrawCardFromEntry( entry, player );
				i = m_distribution.erase( i );
			}

			player = ++player % 2;
		}
		while ( !end_condition_met );
	}

	template< typename TContinueCondition >
	void MulliganIf( TContinueCondition continue_condition, int player = 0 )
	{
		bool end_condition_met;
		do
		{
			end_condition_met = true;

			for ( auto i = m_distribution.begin(); i != m_distribution.end(); )
			{
				const TProbabilityDistributionEntry& entry = *i;
				if ( !continue_condition( entry ) )
				{
					++i;
					continue;
				}

				end_condition_met = false;
				MulliganFromEntry( entry, player );
				i = m_distribution.erase( i );
			}
		}
		while ( !end_condition_met );
	}

	void DrawCard( int player = 0 )
	{
		for ( auto i = m_distribution.begin(); i != m_distribution.end(); ++i )
		{
			TProbabilityDistributionEntry& entry = *i;
			entry.m_dirty = true;
		}
		for ( auto i = m_distribution.begin(); i != m_distribution.end(); )
		{
			const TProbabilityDistributionEntry& entry = *i;
			assert( entry.m_dirty );
			DrawCardFromEntry( entry, player );
			i = m_distribution.erase( i );
		}
	}
	
	void DrawCards( int number_to_draw, int player = 0 )
	{
		for ( int i = 0; i < number_to_draw; ++i )
		{
			DrawCard( player );
		}
	}

	template< typename TMulliganIfCondition >
	void DrawStartingHand( TMulliganIfCondition mulligan_if_condition, int player = 0 )
	{
		DrawCards( 7, player );
		MulliganIf( mulligan_if_condition, player );
	}
	
	void DrawStartingHandFromRules( const CMulliganRules< TCardSetType >& rules, int player = 0 )
	{
		DrawStartingHand(
			[ rules, player ]( const TProbabilityDistributionEntry& entry ) -> bool
			{
				typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType TCardSetType;
				return rules.IsMulliganAdvised( entry.GetPlayer( player ).m_hand.GetSize(), TCardSetType::sGetIndex( entry.GetPlayer( player ).m_hand ) );
			},
			player
		);
	}
	
	template< typename TFunction >
	void ForEach( TFunction function )
	{
		for ( auto i = m_distribution.begin(); i != m_distribution.end(); ++i )
		{
			TProbabilityDistributionEntry& entry = *i;
			function( entry );
		}
	}
		
	template< typename TCondition >
	bool TestIf( TCondition condition )
	{
		for ( auto i = m_distribution.begin(); i != m_distribution.end(); ++i )
		{
			const TProbabilityDistributionEntry& entry = *i;
			if ( !condition( entry ) ) return false;
		}
		return true;
	}

	template< typename TMeasurer >
	double TakeMeasurement( TMeasurer measurer ) const
	{
		double result = 0.0;
		for ( auto i = m_distribution.begin(); i != m_distribution.end(); ++i )
		{
			const TProbabilityDistributionEntry& entry = *i;

			double measurement;
			if ( std::is_same< decltype( measurer( entry ) ), bool >::value )
			{
				measurement = ( measurer( entry ) ? 1.0 : 0.0 );
			}
			else
			{
				measurement = measurer( entry );
			}

			result += entry.m_probability * measurement;
		}
		return result;
	}
	
	template< typename TMeasurer >
	CTakeExtendedMeasurementResults TakeExtendedMeasurement( TMeasurer measurer ) const
	{
		const double average = TakeMeasurement( measurer );

		double variance = 0.0;
		for ( auto i = m_distribution.begin(); i != m_distribution.end(); ++i )
		{
			const TProbabilityDistributionEntry& entry = *i;

			double measurement;
			if ( std::is_same< decltype( measurer( entry ) ), bool >::value )
			{
				measurement = ( measurer( entry ) ? 1.0 : 0.0 );
			}
			else
			{
				measurement = measurer( entry );
			}

			variance += entry.m_probability * ( measurement - average ) * ( measurement - average );
		}
		const double standard_deviation = std::sqrt( variance );

		CTakeExtendedMeasurementResults results = { average, standard_deviation };
		return results;
	}

	template< typename TBinner >
	void BinDistribution( double table[], int table_size, TBinner binner ) const
	{
		std::fill( table, table + table_size, 0.0 );
		for ( auto i = m_distribution.begin(); i != m_distribution.end(); ++i )
		{
			const TProbabilityDistributionEntry& entry = *i;

			const int bin = binner( entry );

			if ( bin >= 0 && bin < table_size )
			{
				table[ bin ] += entry.m_probability;
			}
		}
	}

	void Validate() const
	{
		double total_probability = 0.0;
		for ( auto i = m_distribution.begin(); i != m_distribution.end(); ++i )
		{
			const TProbabilityDistributionEntry& entry = *i;
			assert( !entry.m_dirty );
			total_probability += entry.m_probability;
		}
		assert( total_probability > 0.999 && total_probability < 1.001 );
	}

	void DebugDump() const
	{
		static_assert( TCardSetType::NumberOfCardTypes == 2, "DebugDump not yet supported for arbitrary card sets." );

		Validate();
		for ( auto i = m_distribution.begin(); i != m_distribution.end(); ++i )
		{
			const TProbabilityDistributionEntry& entry = *i;
			if ( TProbabilityDistributionEntry::NumberOfPlayers == 1 )
			{
				printf( "p: %7.3f%%    f:%.1f    ch: %2d    lh: %2d\n", entry.m_probability * 100.0, entry.m_user_data, entry.GetPlayer( 0 ).m_hand.GetSize(), entry.GetPlayer( 0 ).m_hand.m_lands );
			}
			else
			{
				printf( "p: %7.3f%%    f:%.1f    p1[ ch: %2d    lh: %2d ]    p2[ ch: %2d    lh: %2d ]\n", entry.m_probability * 100.0, entry.m_user_data, entry.GetPlayer( 0 ).m_hand.GetSize(), entry.GetPlayer( 0 ).m_hand.m_lands, entry.GetPlayer( 1 ).m_library.m_lands, entry.GetPlayer( 1 ).m_hand.GetSize(), entry.GetPlayer( 1 ).m_hand.m_lands );
			}
		}
		printf( "\n" );
	}
};

typedef CProbabilityDistributionTemplate< CPlayerProbabilityDistributionEntry > CPlayerProbabilityDistribution;
typedef CProbabilityDistributionTemplate< CGameProbabilityDistributionEntry > CGameProbabilityDistribution;
typedef CPlayerProbabilityDistribution CProbabilityDistribution;

template< typename TCardType >
struct CardSetTypes
{
	typedef CPileTemplate< TCardType > CPile;
	typedef CPlayerProbabilityDistributionEntryTemplate< CPlayerTemplate< TCardType > > CProbabilityDistributionEntry;
	typedef CProbabilityDistributionTemplate< CProbabilityDistributionEntry > CProbabilityDistribution;
};

template< typename TType >
struct extract_type_from_lambda_parameter
{
	typedef typename std::remove_reference< typename std::remove_const< TType >::type >::type::TPlayerType::TPileType::TCardSetType TCardSetType;
};

};

#endif // MAGIC_PROBABILITY_TOOLKIT_H