#ifndef MULLIGAN_RULES_H
#define MULLIGAN_RULES_H

#include "magic_probability_toolkit.h"
#include "table_output.h"

namespace MulliganRules
{

auto NeverMulligan = []( const mpt::CProbabilityDistributionEntry& entry )
{
	return false;
};

auto StandardDeck_OnTheDraw = []( const mpt::CProbabilityDistributionEntry& entry ) -> bool
{
	if ( entry.m_player.m_hand.GetSize() == 7 )
	{
		return entry.m_player.m_hand.m_lands < 3 || entry.m_player.m_hand.m_lands > 5;
	}
	if ( entry.m_player.m_hand.GetSize() == 6 )
	{
		return entry.m_player.m_hand.m_lands < 2 || entry.m_player.m_hand.m_lands > 5;
	}
	if ( entry.m_player.m_hand.GetSize() == 5 )
	{
		return entry.m_player.m_hand.m_lands < 2 || entry.m_player.m_hand.m_lands > 4;
	}
	return false;
};

auto StandardDeck_OnThePlay = []( const mpt::CProbabilityDistributionEntry& entry ) -> bool
{
	if ( entry.m_player.m_hand.GetSize() == 7 )
	{
		return entry.m_player.m_hand.m_lands < 3 || entry.m_player.m_hand.m_lands > 5;
	}
	if ( entry.m_player.m_hand.GetSize() == 6 )
	{
		return entry.m_player.m_hand.m_lands < 2 || entry.m_player.m_hand.m_lands > 4;
	}
	if ( entry.m_player.m_hand.GetSize() == 5 )
	{
		return entry.m_player.m_hand.m_lands < 1 || entry.m_player.m_hand.m_lands > 4;
	}
	return false;
};

}

namespace mlr
{

enum EComplexMulliganAction
{
	Mulligan,
	ConsiderMulligan,
	Keep
};

struct CCalculateMulliganRulesResults
{
	typedef mpt::CMulliganRules< mpt::CardSet::IgnoringColor > TRulesType;

	TRulesType m_rules;

	EComplexMulliganAction m_complex_mulligan_rules_table[ mpt::HandSizePlusOne ][ mpt::HandSizePlusOne ];

	double m_chance_of_drawing_hand_table[ mpt::HandSizePlusOne ][ mpt::HandSizePlusOne ];
	double m_probability_of_mulligan[ mpt::HandSizePlusOne ];

	double m_chance_of_competitive_start_table[ mpt::HandSizePlusOne ][ mpt::HandSizePlusOne ];
	double m_overall_chance_of_competitive_start_table[ mpt::HandSizePlusOne ];
	double m_unadjusted_chance_of_competitive_start[ mpt::HandSizePlusOne ];

	double m_chance_of_secondary_criteria_table[ mpt::HandSizePlusOne ][ mpt::HandSizePlusOne ];
	double m_overall_chance_of_secondary_criteria_table[ mpt::HandSizePlusOne ];
	int m_hand_size_secondary_criteria_used_for;

	int m_first_interesting_hand_size;

	bool IsSecondaryCriteriaUsed() { return m_hand_size_secondary_criteria_used_for != -1; }

	void PrintTable( const char* header, int min_hand_size = 1 );
	void PrintComplexTable( const char* header );
	void PrintEntireCalculation( const char* header );
	void PrintCleanTable( const char* header );
};

template< typename TCompetitiveStartMeasurement >
CCalculateMulliganRulesResults CalculateMulliganRulesIgnoringColor( mpt::TPlayOrDraw play_or_draw, int turn_of_competitive_start_measurement, TCompetitiveStartMeasurement competitive_start_measurement, int deck_size = 40, int land_count = 17 )
{
	CCalculateMulliganRulesResults results;

	results.m_overall_chance_of_competitive_start_table[ 0 ] = 0.0;
	results.m_overall_chance_of_secondary_criteria_table[ 0 ] = 0.0;
	results.m_hand_size_secondary_criteria_used_for = -1;
	results.m_first_interesting_hand_size = -1;

	for ( int starting_hand_size = 1; starting_hand_size < 8; ++starting_hand_size )
	{
		results.m_probability_of_mulligan[ starting_hand_size ] = 0.0;

		for ( int lands_in_starting_hand = 0; lands_in_starting_hand <= starting_hand_size; ++lands_in_starting_hand )
		{
			CProbabilityDistribution state( deck_size, land_count, starting_hand_size, lands_in_starting_hand );

			const int draws_before_measurement = play_or_draw == Draw ? turn_of_competitive_start_measurement : turn_of_competitive_start_measurement - 1;
			state.DrawCards( draws_before_measurement );

			results.m_chance_of_competitive_start_table[ starting_hand_size ][ lands_in_starting_hand ] = state.TakeMeasurement( competitive_start_measurement );

			state.DrawCard();
			results.m_chance_of_secondary_criteria_table[ starting_hand_size ][ lands_in_starting_hand ] = state.TakeMeasurement( competitive_start_measurement );
		}

		{
			CProbabilityDistribution state( deck_size, land_count );
			state.DrawCards( starting_hand_size );
			state.BinDistribution(
				results.m_chance_of_drawing_hand_table[ starting_hand_size ],
				8,
				[]( const CProbabilityDistributionEntry& entry )
				{
					return entry.m_player.m_hand.m_lands;
				}
			);
		}

		{
			double mulligan_adjusted_chance_of_competitive_start = 0.0;
			double unadjusted_chance_of_competitive_start = 0.0;

			double mulligan_adjusted_chance_of_secondary_criteria = 0.0;

			for ( int lands_in_starting_hand = 0; lands_in_starting_hand <= starting_hand_size; ++lands_in_starting_hand )
			{
				if ( results.m_chance_of_competitive_start_table[ starting_hand_size ][ lands_in_starting_hand ] > 0.0 || results.m_overall_chance_of_competitive_start_table[ starting_hand_size - 1 ] > 0.0 )
				{
					results.m_rules.m_table[ starting_hand_size ][ lands_in_starting_hand ] = results.m_chance_of_competitive_start_table[ starting_hand_size ][ lands_in_starting_hand ] < results.m_overall_chance_of_competitive_start_table[ starting_hand_size - 1 ];
				}
				else
				{
					results.m_rules.m_table[ starting_hand_size ][ lands_in_starting_hand ] = results.m_chance_of_secondary_criteria_table[ starting_hand_size ][ lands_in_starting_hand ] < results.m_overall_chance_of_secondary_criteria_table[ starting_hand_size - 1 ];
					results.m_hand_size_secondary_criteria_used_for = starting_hand_size;
				}

				const double larger = std::max( results.m_chance_of_competitive_start_table[ starting_hand_size ][ lands_in_starting_hand ], results.m_overall_chance_of_competitive_start_table[ starting_hand_size - 1 ] );
				if ( abs( results.m_chance_of_competitive_start_table[ starting_hand_size ][ lands_in_starting_hand ] - results.m_overall_chance_of_competitive_start_table[ starting_hand_size - 1 ] ) < 0.125 * larger )
				{
					results.m_complex_mulligan_rules_table[ starting_hand_size ][ lands_in_starting_hand ] = ConsiderMulligan;
				}
				else if ( results.m_rules.m_table[ starting_hand_size ][ lands_in_starting_hand ] )
				{
					results.m_complex_mulligan_rules_table[ starting_hand_size ][ lands_in_starting_hand ] = Mulligan;
				}
				else
				{
					results.m_complex_mulligan_rules_table[ starting_hand_size ][ lands_in_starting_hand ] = Keep;
				}

				if ( results.m_rules.m_table[ starting_hand_size ][ lands_in_starting_hand ] )
				{
					results.m_probability_of_mulligan[ starting_hand_size ] += results.m_chance_of_drawing_hand_table[ starting_hand_size ][ lands_in_starting_hand ];
				}

				{
					const double chance_of_competitive_start = results.m_rules.m_table[ starting_hand_size ][ lands_in_starting_hand ]
						? results.m_overall_chance_of_competitive_start_table[ starting_hand_size - 1 ]
						: results.m_chance_of_competitive_start_table[ starting_hand_size ][ lands_in_starting_hand ];
					const double contributing_chance_of_competitive_start = results.m_chance_of_drawing_hand_table[ starting_hand_size ][ lands_in_starting_hand ] * chance_of_competitive_start;
					mulligan_adjusted_chance_of_competitive_start += contributing_chance_of_competitive_start;
				}

				{
					const double chance_of_secondary_criteria = results.m_rules.m_table[ starting_hand_size ][ lands_in_starting_hand ]
						? results.m_overall_chance_of_secondary_criteria_table[ starting_hand_size - 1 ]
						: results.m_chance_of_secondary_criteria_table[ starting_hand_size ][ lands_in_starting_hand ];
					const double contributing_chance_of_secondary_criteria = results.m_chance_of_drawing_hand_table[ starting_hand_size ][ lands_in_starting_hand ] * chance_of_secondary_criteria;
					mulligan_adjusted_chance_of_secondary_criteria += contributing_chance_of_secondary_criteria;
				}

				unadjusted_chance_of_competitive_start += results.m_chance_of_drawing_hand_table[ starting_hand_size ][ lands_in_starting_hand ] * results.m_chance_of_competitive_start_table[ starting_hand_size ][ lands_in_starting_hand ];
			}

			results.m_overall_chance_of_competitive_start_table[ starting_hand_size ] = mulligan_adjusted_chance_of_competitive_start;
			results.m_unadjusted_chance_of_competitive_start[ starting_hand_size ] = unadjusted_chance_of_competitive_start;

			results.m_overall_chance_of_secondary_criteria_table[ starting_hand_size ] = mulligan_adjusted_chance_of_secondary_criteria;

			if ( results.m_first_interesting_hand_size == -1 && results.m_overall_chance_of_competitive_start_table[ starting_hand_size ] > 0.0 )
			{
				results.m_first_interesting_hand_size = starting_hand_size;
			}
		}
	}

	return results;
}

template< typename TProbabilityDistributionType, typename TInitialHandEvaluation, typename TCompetitiveStartMeasurement >
mpt::CMulliganRules< typename TProbabilityDistributionType::TCardSetType > CalculateMulliganRules( mpt::TPlayOrDraw play_or_draw, int turn_of_competitive_start_measurement, TInitialHandEvaluation initial_hand_evaluation, TCompetitiveStartMeasurement competitive_start_measurement, const typename TProbabilityDistributionType::TPileType& deck )
{
	typedef TProbabilityDistributionType::TCardSetType TCardSetType;
	typedef mpt::CMulliganRules< TCardSetType > TRulesType;

	double chance_of_drawing_hand_table[ mpt::HandSizePlusOne ][ TCardSetType::NumberOfHandsPerHandSize ];
	double chance_of_competitive_start_table[ mpt::HandSizePlusOne ][ TCardSetType::NumberOfHandsPerHandSize ];
	double overall_chance_of_competitive_start_table[ mpt::HandSizePlusOne ];
	double chance_of_secondary_criteria_table[ mpt::HandSizePlusOne ][ TCardSetType::NumberOfHandsPerHandSize ];
	double overall_chance_of_secondary_criteria_table[ mpt::HandSizePlusOne ];

	mpt::CMulliganRules< TProbabilityDistributionType::TCardSetType > rules;

	overall_chance_of_competitive_start_table[ 0 ] = 0.0;
	overall_chance_of_secondary_criteria_table[ 0 ] = 0.0;

	for ( int starting_hand_size = 1; starting_hand_size < 8; ++starting_hand_size )
	{
		for ( int hand_index = 0; !TCardSetType::sIsBeyondLastHand( starting_hand_size, hand_index ); ++hand_index )
		{
			TProbabilityDistributionType::TPileType hand, library;
			if ( !TCardSetType::sGetHandAndLibrary( hand, library, deck, starting_hand_size, hand_index ) ) continue;

			TProbabilityDistributionType state( library, hand );

			state.ForEach( initial_hand_evaluation );

			if ( play_or_draw == Draw )
			{
				state.DrawCard();
			}

			state.DrawCards( turn_of_competitive_start_measurement - 1 );

			chance_of_competitive_start_table[ starting_hand_size ][ hand_index ] = state.TakeMeasurement( competitive_start_measurement );

			state.DrawCard();
			chance_of_secondary_criteria_table[ starting_hand_size ][ hand_index ] = state.TakeMeasurement( competitive_start_measurement );
		}

		{
			TProbabilityDistributionType state( deck );
			state.DrawCards( starting_hand_size );
			state.BinDistribution(
				chance_of_drawing_hand_table[ starting_hand_size ],
				TCardSetType::NumberOfHandsPerHandSize,
				[]( const TProbabilityDistributionType::TProbabilityDistributionEntry& entry ) -> int
				{
					typedef extract_type_from_lambda_parameter< decltype( entry ) >::TCardSetType TCardSetType;

					return TCardSetType::sGetIndex( entry.m_player.m_hand );
				}
			);
		}

		{
			double mulligan_adjusted_chance_of_competitive_start = 0.0;
			double mulligan_adjusted_chance_of_secondary_criteria = 0.0;
			for ( int hand_index = 0; !TCardSetType::sIsBeyondLastHand( starting_hand_size, hand_index ); ++hand_index )
			{
				TProbabilityDistributionType::TPileType hand, library;
				if ( !TCardSetType::sGetHandAndLibrary( hand, library, deck, starting_hand_size, hand_index ) ) continue;

				if ( chance_of_competitive_start_table[ starting_hand_size ][ hand_index ] > 0.0 || overall_chance_of_competitive_start_table[ starting_hand_size - 1 ] > 0.0 )
				{
					rules.m_table[ starting_hand_size ][ hand_index ] = chance_of_competitive_start_table[ starting_hand_size ][ hand_index ] < overall_chance_of_competitive_start_table[ starting_hand_size - 1 ];
				}
				else
				{
					rules.m_table[ starting_hand_size ][ hand_index ] = chance_of_secondary_criteria_table[ starting_hand_size ][ hand_index ] < overall_chance_of_secondary_criteria_table[ starting_hand_size - 1 ];
				}

				{
					const double chance_of_competitive_start = rules.m_table[ starting_hand_size ][ hand_index ]
						? overall_chance_of_competitive_start_table[ starting_hand_size - 1 ]
						: chance_of_competitive_start_table[ starting_hand_size ][ hand_index ];
					const double contributing_chance_of_competitive_start = chance_of_drawing_hand_table[ starting_hand_size ][ hand_index ] * chance_of_competitive_start;
					mulligan_adjusted_chance_of_competitive_start += contributing_chance_of_competitive_start;
				}

				{
					const double chance_of_secondary_criteria = rules.m_table[ starting_hand_size ][ hand_index ]
						? overall_chance_of_secondary_criteria_table[ starting_hand_size - 1 ]
						: chance_of_secondary_criteria_table[ starting_hand_size ][ hand_index ];
					const double contributing_chance_of_secondary_criteria = chance_of_drawing_hand_table[ starting_hand_size ][ hand_index ] * chance_of_secondary_criteria;
					mulligan_adjusted_chance_of_secondary_criteria += contributing_chance_of_secondary_criteria;
				}
			}

			overall_chance_of_competitive_start_table[ starting_hand_size ] = mulligan_adjusted_chance_of_competitive_start;
			overall_chance_of_secondary_criteria_table[ starting_hand_size ] = mulligan_adjusted_chance_of_secondary_criteria;
		}
	}

	return rules;
}

template< typename TProbabilityDistributionType, typename TInitialHandEvaluation, typename TCompetitiveStartMeasurement >
mpt::CMulliganRulesSet< typename TProbabilityDistributionType::TCardSetType > CalculateMulliganRulesSet( const typename TProbabilityDistributionType::TPileType& deck, TInitialHandEvaluation initial_hand_evaluation, TCompetitiveStartMeasurement competitive_start_measurement )
{
	mpt::CMulliganRulesSet< TProbabilityDistributionType::TCardSetType > rules_set;

	rules_set.m_on_the_play = CalculateMulliganRules<
		TProbabilityDistributionType,
		decltype( initial_hand_evaluation ),
		decltype( competitive_start_measurement )
	>( Play, 4, initial_hand_evaluation, competitive_start_measurement, deck );

	rules_set.m_on_the_draw = CalculateMulliganRules<
		TProbabilityDistributionType,
		decltype( initial_hand_evaluation ),
		decltype( competitive_start_measurement )
	>( Draw, 4, initial_hand_evaluation, competitive_start_measurement, deck );

	return rules_set;
}

}

#endif // MULLIGAN_RULES_H