#ifndef TABLE_OUTPUT_H
#define TABLE_OUTPUT_H

#include <cassert>
#include <cstdarg>
#include <list>

namespace tbo
{

struct CTable
{
	static const int header_width = 77;
	static const int column_separation = 3;
	static const int buffer_size = 4096;

	char m_character_buffer[ buffer_size ];
	int m_next_character_index;

	std::list< int > m_column_header_list;
	std::list< int > m_entry_list;
	std::list< int > m_addendum_header_list;

	enum EOutputSetting
	{
		HumanReadable,
		CSV,
		NumberOfOutputSettings
	};
	static EOutputSetting sm_output_setting;

	CTable( const char* header, ... )
	{
		va_list args;
		va_start( args, header );
		m_next_character_index = 1 + vsprintf_s( m_character_buffer, header, args );
		va_end( args );
	}

	static void sSetOutputSetting( EOutputSetting setting )
	{
		assert( setting < NumberOfOutputSettings );
		sm_output_setting = setting;
	}
	
	static EOutputSetting sGetOutputSetting()
	{
		return sm_output_setting;
	}

	void AddColumn( const char* column, ... )
	{
		va_list args;
		va_start( args, column );

		m_column_header_list.push_back( m_next_character_index );
		if ( sm_output_setting == HumanReadable )
		{
			m_next_character_index += 1 + vsprintf_s( &m_character_buffer[ m_next_character_index ], buffer_size - m_next_character_index, column, args );
		}
		else
		{
			m_character_buffer[ m_next_character_index++ ] = '\"';
			m_next_character_index += vsprintf_s( &m_character_buffer[ m_next_character_index ], buffer_size - m_next_character_index, column, args );
			m_character_buffer[ m_next_character_index++ ] = '\"';
			m_character_buffer[ m_next_character_index++ ] = '\0';
		}

		va_end( args );
	}

	void AddAddendum( const char* addendum )
	{
		m_addendum_header_list.push_back( m_next_character_index );
		if ( sm_output_setting == HumanReadable )
		{
			m_next_character_index += 1 + sprintf_s( &m_character_buffer[ m_next_character_index ], buffer_size - m_next_character_index, "%s:", addendum );
		}
		else
		{
			m_next_character_index += 1 + sprintf_s( &m_character_buffer[ m_next_character_index ], buffer_size - m_next_character_index, "\"%s\"", addendum );
		}
	}

	void AddInteger( int data )
	{
		m_entry_list.push_back( m_next_character_index );
		if ( sm_output_setting == HumanReadable )
		{
			m_next_character_index += 1 + sprintf_s( &m_character_buffer[ m_next_character_index ], buffer_size - m_next_character_index, "%2d", data );
		}
		else
		{
			m_next_character_index += 1 + sprintf_s( &m_character_buffer[ m_next_character_index ], buffer_size - m_next_character_index, "%d", data );
		}
	}

	void AddNumber( double data, int decimals = 1 )
	{
		m_entry_list.push_back( m_next_character_index );
		if ( sm_output_setting == HumanReadable )
		{
			m_next_character_index += 1 + sprintf_s( &m_character_buffer[ m_next_character_index ], buffer_size - m_next_character_index, "%4.*f", decimals, data );
		}
		else
		{
			m_next_character_index += 1 + sprintf_s( &m_character_buffer[ m_next_character_index ], buffer_size - m_next_character_index, "%f", data );
		}
	}

	void AddProbability( double data, int decimals = 1, bool blank_if_zero = false )
	{
		m_entry_list.push_back( m_next_character_index );
		if ( sm_output_setting == HumanReadable )
		{
			if ( blank_if_zero && data == 0.0 )
			{
				m_next_character_index += 1 + sprintf_s( &m_character_buffer[ m_next_character_index ], buffer_size - m_next_character_index, "" );
			}
			else
			{
				m_next_character_index += 1 + sprintf_s( &m_character_buffer[ m_next_character_index ], buffer_size - m_next_character_index, "%5.*f%%", decimals, data * 100.0 );
			}
		}
		else
		{
			m_next_character_index += 1 + sprintf_s( &m_character_buffer[ m_next_character_index ], buffer_size - m_next_character_index, "%.10f", data );
		}
	}

	void AddString( const char* data, ... )
	{
		va_list args;
		va_start( args, data );

		m_entry_list.push_back( m_next_character_index );
		if ( sm_output_setting == HumanReadable )
		{
			m_next_character_index += 1 + vsprintf_s( &m_character_buffer[ m_next_character_index ], buffer_size - m_next_character_index, data, args );
		}
		else
		{
			m_character_buffer[ m_next_character_index++ ] = '\"';
			m_next_character_index += vsprintf_s( &m_character_buffer[ m_next_character_index ], buffer_size - m_next_character_index, data, args );
			m_character_buffer[ m_next_character_index++ ] = '\"';
			m_character_buffer[ m_next_character_index++ ] = '\0';
		}

		va_end( args );
	}

	static void PrintLine( char symbol )
	{
		printf( "+" );
		for ( int count = 0; count < header_width; ++count )
		{
			printf( "%c", symbol );
		}
		printf( "+\n" );
	}
	
	void PrintHeader() const
	{
		printf( "\n" );
		PrintLine( '=' );
		const char* next_character = m_character_buffer;
		while ( *next_character )
		{
			printf( " " );
			const char* const line_start = next_character;
			while ( *next_character && next_character - line_start < header_width )
			{
				++next_character;
			}
			if ( *next_character )
			{
				while ( *next_character != ' ' )
				{
					--next_character;
					assert( next_character > line_start );
				}
			}
			printf( "%.*s\n", next_character - line_start, line_start );
			while ( *next_character && *next_character == ' ' )
			{
				++next_character;
			}
		}
		PrintLine( '-' );
	}

	void PrintHumanReadable() const
	{
		PrintHeader();

		const int number_of_columns = m_column_header_list.size();
		const int number_of_addendums = m_addendum_header_list.size();
		const int number_of_entries = m_entry_list.size();
		assert( number_of_entries >= number_of_addendums );

		auto e = m_entry_list.begin();

		if ( number_of_columns > 0 )
		{
			assert( ( number_of_entries - number_of_addendums ) % number_of_columns == 0 );
			const int number_of_rows = ( number_of_entries - number_of_addendums ) / number_of_columns;

			std::list< size_t > column_width_list;
			for ( auto c = m_column_header_list.begin(); c != m_column_header_list.end(); ++c )
			{
				column_width_list.push_back( strlen( &m_character_buffer[ *c ] ) );
			}
			for ( int row = 0; row < number_of_rows; ++row )
			{
				for ( auto w = column_width_list.begin(); w != column_width_list.end(); ++e, ++w )
				{
					*w = std::max( *w, strlen( &m_character_buffer[ *e ] ) );
				}
			}

			printf( " " );
			auto c = m_column_header_list.begin();
			for ( auto w = column_width_list.begin(); w != column_width_list.end(); ++c, ++w )
			{
				printf( "%-*s", *w + column_separation, &m_character_buffer[ *c ] );
			}
			printf( "\n" );

			e = m_entry_list.begin();
			for ( int row = 0; row < number_of_rows; ++row )
			{
				printf( " " );
				for ( auto w = column_width_list.begin(); w != column_width_list.end(); ++e, ++w )
				{
					printf( " %-*s", *w + column_separation - 1, &m_character_buffer[ *e ] );
				}
				printf( "\n" );
			}

			if ( number_of_addendums > 0 )
			{
				printf( "\n" );
			}
		}

		if ( number_of_addendums > 0 )
		{
			size_t addendum_width = 0;
			for ( auto a = m_addendum_header_list.begin(); a != m_addendum_header_list.end(); ++a )
			{
				addendum_width = std::max( addendum_width, strlen( &m_character_buffer[ *a ] ) );
			}

			for ( auto a = m_addendum_header_list.begin(); a != m_addendum_header_list.end(); ++e, ++a )
			{
				printf( "  %-*s %s\n", addendum_width, &m_character_buffer[ *a ], &m_character_buffer[ *e ] );
			}
		}
	}

	void PrintCSV() const
	{
		printf( "\n" );
		printf( "\"%s\"\n", m_character_buffer );

		const int number_of_columns = m_column_header_list.size();
		const int number_of_addendums = m_addendum_header_list.size();
		const int number_of_entries = m_entry_list.size();
		assert( number_of_entries >= number_of_addendums );

		auto e = m_entry_list.begin();

		if ( number_of_columns > 0 )
		{
			assert( ( number_of_entries - number_of_addendums ) % number_of_columns == 0 );
			const int number_of_rows = ( number_of_entries - number_of_addendums ) / number_of_columns;

			for ( auto c = m_column_header_list.begin(); c != m_column_header_list.end(); ++c )
			{
				printf( "%s,", &m_character_buffer[ *c ] );
			}
			printf( "\n" );
			for ( int row = 0; row < number_of_rows; ++row )
			{
				for ( int column = 0; column < number_of_columns; ++e, ++column )
				{
					printf( "%s,", &m_character_buffer[ *e ] );
				}
				printf( "\n" );
			}

			if ( number_of_addendums > 0 )
			{
				printf( "\n" );
			}
		}

		if ( number_of_addendums > 0 )
		{
			for ( auto a = m_addendum_header_list.begin(); a != m_addendum_header_list.end(); ++e, ++a )
			{
				printf( "%s,%s\n", &m_character_buffer[ *a ], &m_character_buffer[ *e ] );
			}
		}
	}

	void Print() const
	{
		if ( sm_output_setting == HumanReadable )
		{
			PrintHumanReadable();
		}
		else if ( sm_output_setting == CSV )
		{
			PrintCSV();
		}
		fflush( stdout );
	}
};

}

#endif // TABLE_OUTPUT_H